/*
 * $Date: 1998/05/11 15:37:17 $ $Id: HCDisk.c,v 1.13 1998/05/11 15:37:17 dahlin Exp $
 *
 * AUTHOR: Mike Dahlin, UT Austin
 *
 * SQUID Internet Object Cache  http://squid.nlanr.net/Squid/
 * --------------------------------------------------------
 *
 *  Squid is the result of efforts by numerous individuals from the
 *  Internet community.  Development is led by Duane Wessels of the
 *  National Laboratory for Applied Network Research and funded by
 *  the National Science Foundation.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *  
 */

/*
 * Copyright (c) 1994, 1995.  All rights reserved.
 *  
 *   The Harvest software was developed by the Internet Research Task
 *   Force Research Group on Resource Discovery (IRTF-RD):
 *  
 *         Mic Bowman of Transarc Corporation.
 *         Peter Danzig of the University of Southern California.
 *         Darren R. Hardy of the University of Colorado at Boulder.
 *         Udi Manber of the University of Arizona.
 *         Michael F. Schwartz of the University of Colorado at Boulder.
 *         Duane Wessels of the University of Colorado at Boulder.
 *  
 *   This copyright notice applies to software in the Harvest
 *   ``src/'' directory only.  Users should consult the individual
 *   copyright notices in the ``components/'' subdirectories for
 *   copyright information about other software bundled with the
 *   Harvest source code distribution.
 *  
 * TERMS OF USE
 *   
 *   The Harvest software may be used and re-distributed without
 *   charge, provided that the software origin and research team are
 *   cited in any use of the system.  Most commonly this is
 *   accomplished by including a link to the Harvest Home Page
 *   (http://harvest.cs.colorado.edu/) from the query page of any
 *   Broker you deploy, as well as in the query result pages.  These
 *   links are generated automatically by the standard Broker
 *   software distribution.
 *   
 *   The Harvest software is provided ``as is'', without express or
 *   implied warranty, and with no support nor obligation to assist
 *   in its use, correction, modification or enhancement.  We assume
 *   no liability with respect to the infringement of copyrights,
 *   trade secrets, or any patents, and are not responsible for
 *   consequential damages.  Proper use of the Harvest software is
 *   entirely the responsibility of the user.
 *  
 * DERIVATIVE WORKS
 *  
 *   Users may make derivative works from the Harvest software, subject 
 *   to the following constraints:
 *  
 *     - You must include the above copyright notice and these 
 *       accompanying paragraphs in all forms of derivative works, 
 *       and any documentation and other materials related to such 
 *       distribution and use acknowledge that the software was 
 *       developed at the above institutions.
 *  
 *     - You must notify IRTF-RD regarding your distribution of 
 *       the derivative work.
 *  
 *     - You must clearly notify users that your are distributing 
 *       a modified version and not the original Harvest software.
 *  
 *     - Any derivative product is also subject to these copyright 
 *       and use restrictions.
 *  
 *   Note that the Harvest software is NOT in the public domain.  We
 *   retain copyright, as specified above.
 *  
 * HISTORY OF FREE SOFTWARE STATUS
 *  
 *   Originally we required sites to license the software in cases
 *   where they were going to build commercial products/services
 *   around Harvest.  In June 1995 we changed this policy.  We now
 *   allow people to use the core Harvest software (the code found in
 *   the Harvest ``src/'' directory) for free.  We made this change
 *   in the interest of encouraging the widest possible deployment of
 *   the technology.  The Harvest software is really a reference
 *   implementation of a set of protocols and formats, some of which
 *   we intend to standardize.  We encourage commercial
 *   re-implementations of code complying to this set of standards.  
 */
/*
 *------------------------------------------------------------------
 *
 * LESS Group 
 * Computer Sciences Department 
 * University of Texas at Austin
 *
 * HCDisk ---
 *          The on-disk cache of hints. The key data structure
 *          is an mmapped array which is treated as a <entriesPerBucket>-
 *          associative cache.
 *
 * Copyright (c) 1997 by Mike Dahlin.
 * 
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for noncommercial use without fee and
 * without written agreement is hereby granted, provided that 
 * the above copyright notice appear in all copies of this software.
 * 
 *------------------------------------------------------------------
 */
#include "squid.h"
#include <assert.h>
#include <sys/mman.h>

#define BIG_ENOUGH 1024


static const char *suffix = ".theCache";

static int readLine(int fd, char *buffer, int max);
static void initializeFile(int dataFD, int nbuckets, int associativity);
static void init(HCDisk *d, char *diskPath);
static int findKey(HCDisk *d, HCEntry *bucketp, URLKey key, 
		   HCEntry *foundRet);
static int deleteKey(HCDisk *d, HCEntry *bp, HCEntry *oldEntry, 
		     HCEntry *survivingEntry);
static void insertKey(HCDisk *d, HCEntry *bp, HCEntry *newEntry);
static caddr_t pageAddr(caddr_t e);
static void sanityCheckMatch(HCDisk *d, HCEntry *bucketp, HCEntry *entry, 
			     int expectFullMatch, int expectKeyMatch);
static HCEntry *bucket(HCDisk *d, URLKey key);
#ifdef DOTEST
static void timeMadvise();
#endif


static long pageSize;

/*
 * Should we try to keep stuff we just read or wrote in the cache.
 */
/*
 * You may not want to cache writes since they may or may not
 * be about stuff you will later read. On the other hand, there
 * may be temporal locality about stuff that gets updated --
 * I may be likely to read or discard the object in the near
 * future. Also, telling system to kick it out of cache costs
 * about 1ms per call.
 */
static const int CACHE_RECENT_WRITES = 1;
/*
 * You may not want to cache reads since these are things
 * that you will now have in your cache (so why would
 * you look at the hints any time soon. On the other hand,
 * telling system to kick out reads costs about 1ms per call
 */
static const int CACHE_RECENT_READS = 0;
/* 
 * You may not want to prefetch at all (we can only prefetch
 * on network updates, and those are not on critical
 * path anywhere. Maybe better to sacrifice latency for
 * reduced overhead.
 * 
 * No point int trying to pipeline if there are not many
 * updates to pipeline. Otherwise the overhead of the
 * system calls outweighs the reduced latency for disk
 * reads. In the best case, madvise (tight loop of madvise
 * about same location), madvise appears to take about 1ms per call.
 */
static const int DO_PREFETCH = 0;

/*
 *------------------------------------------------------------------
 *
 * HCDisk_Init --
 *
 *          Constructor
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
void 
HCDisk_Init(HCDisk *d, char *diskPath)
{
  if((pageSize = sysconf(_SC_PAGESIZE)) < 0){
    perror("Cannot determine page size\n");
    exit(-1);
  }
  init(d, diskPath);  /* sets d->nbuckets, entriesPerBucket */
}

/*
 *------------------------------------------------------------------
 *
 * init --
 *
 *          Does teh set-up work
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *---------------------------------------------------4.22.1998------
 */
static void
init(HCDisk *d, char *diskPath)
{
  int header;
  int dataF;
  char buffer[BIG_ENOUGH], token[BIG_ENOUGH], dataFile[BIG_ENOUGH];
  int version;
  int size, buckets, associativity;
  int gotVersion = 0, gotData = 0, gotAssoc = 0, 
    gotSize = 0, gotBuckets = 0;

  header = open(diskPath, O_RDONLY);
  if(header < 0){
    perror("Opening Hint Cache header");
    exit(-1);
  }

  while(readLine(header, buffer, BIG_ENOUGH)){
    if(buffer[0] != '#'){
      sscanf(buffer, "%s", token);
      if(!strcmp(token, "VERSION")){
	gotVersion = sscanf(buffer, "VERSION %d\n", &version);
	assert(gotVersion == 1);
	assert(version == HCD_VERSION);
      }
      else if(!strcmp(token, "DATA_FILE")){
	gotData = sscanf(buffer, "DATA_FILE %s\n", dataFile);
	assert(gotData == 1);
      }
      else if(!strcmp(token, "SIZE_BYTES")){
	gotSize = sscanf(buffer, "SIZE_BYTES %d\n", &size);
	assert(gotSize == 1);
      }
      else if(!strcmp(token, "BUCKETS")){
	gotBuckets = sscanf(buffer, "BUCKETS %d\n", &buckets);
	assert(gotBuckets == 1);
      }
      else if(!strcmp(token, "ASSOCIATIVITY")){
	gotAssoc = sscanf(buffer, "ASSOCIATIVITY %d\n", &associativity);
	assert(gotAssoc == 1);
      }
      else{
	/* Ignore what we don't understand */
      }
    }
  }
  close(header);
  assert(gotVersion && gotData && gotAssoc && gotSize && gotBuckets);
  assert(size == buckets * associativity * sizeof(HCEntry));
  d->nbuckets = buckets;
  d->entriesPerBucket = associativity;
  dataF =  open(dataFile, O_RDWR);
  if(dataF < 0){
    perror("Opening Hint Cache data");
    exit(-1);
  }
  d->mmappedArray = (HCEntry *)mmap(0, size, PROT_READ|PROT_WRITE,
				   MAP_SHARED, dataF, 0);
  if(!d->mmappedArray){
    perror("MMapping hint cache data");
    exit(-1);
  }
  if(madvise((char *)d->mmappedArray, size, MADV_RANDOM)){
    perror("WARNING: Madvise on hint cache data failed");
  }    
  close(dataF);
}


/*
 *------------------------------------------------------------------
 *
 * readLine --
 *
 *          Read one line of input from file. We really want
 *          to use fgets(), but the fopen() etc. routines
 *          don't work if you have large numbers of
 *          files open (60 or 255 are the max for some implementations).
 *          So, we use the basic open/read/close routines
 *          instead, which let us open up to the max number
 *          of allowed file descriptors.
 *
 *          Store a line of text ending with \n\0 in buffer
 *          and return nonzero if we succeeded in getting
 *          a well-formed line.
 *
 * Arguments:
 *      None.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------4.22.1998-------
 */
static int
readLine(int fd, char *buffer, int max)
{
  ssize_t got;
  int count = 0;

  assert(buffer);
  assert(fd >= 0);
  assert(fd <= 10000);
  assert(max > 2); /* need room for \n\0 */

  while(count < max-2){
    got = read(fd, &buffer[count], 1);
    if(got != 1){
      return 0;
    }
    if(buffer[count] == '\n'){
      buffer[count+1] = '\0';
      return 1;
    }
    count++;
  }
  return 0;
}

/*
 *------------------------------------------------------------------
 *
 * HCDisk_CreateFile --
 *
 *          Does the hard work of setting up a new empty file.
 *
 *          Note: we use raw open() and write() rather than
 *          fopen() and fwrite() because the f-routines
 *          get unhappy in programs that open lots of files
 *          (they often die if you have more than 60 or 255
 *          files open, even if you are allowed to 
 *          have more open file descriptors than that).
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *---------------------------------------------------4.22.1998----------
 */
void 
HCDisk_CreateFile(char *hcPath, int size, int associativity)
{
  char *dataPath;
  int buckets;
  int headerF, dataF;
  char buffer[BIG_ENOUGH];

  dataPath = (char *)malloc(strlen(hcPath) + strlen(suffix) + 1);
  assert(dataPath);
  sprintf(dataPath, "%s%s", hcPath, suffix);

  buckets = (size / associativity) / sizeof(HCEntry);
  assert(buckets * associativity * sizeof(HCEntry) <= size);
  size = buckets * associativity * sizeof(HCEntry);

  headerF = open(hcPath, O_WRONLY|O_CREAT|O_TRUNC, 0744);
  if(headerF < 0){
    perror("Opening Hint Cache header");
    exit(-1);
  }
  sprintf(buffer, "# *** Header for hint cache. Do not modify ***\n");
  assert(write(headerF, buffer, strlen(buffer)) == strlen(buffer));
  sprintf(buffer, "VERSION %d\n", HCD_VERSION);
  assert(write(headerF, buffer, strlen(buffer)) == strlen(buffer));
  sprintf(buffer, "DATA_FILE %s\n", dataPath);
  assert(write(headerF, buffer, strlen(buffer)) == strlen(buffer));
  sprintf(buffer, "SIZE_BYTES %d\n", size);
  assert(write(headerF, buffer, strlen(buffer)) == strlen(buffer));
  sprintf(buffer, "BUCKETS %d\n", buckets);
  assert(write(headerF, buffer, strlen(buffer)) == strlen(buffer));
  sprintf(buffer, "ASSOCIATIVITY %d\n", associativity);
  assert(write(headerF, buffer, strlen(buffer)) == strlen(buffer));

  close(headerF);
  
  dataF = open(dataPath, O_WRONLY|O_CREAT|O_TRUNC, 0744);
  if(dataF < 0){
    perror("Opening Hint Cache data");
    exit(-1);
  }
  initializeFile(dataF, buckets, associativity);
  close(dataF);
  free(dataPath);
}

/*
 *------------------------------------------------------------------
 *
 * initializeFile --
 *
 *          Set a file to empty by invalidating all entries.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
static void
initializeFile(int fd, int nbuckets, int associativity)
{
  HCEntry invalidEntry;
  int ibucket, ientry;
  int count;
  
  HCE_INVALIDATE(&invalidEntry);

  for(ibucket = 0; ibucket < nbuckets; ibucket++){
    for(ientry = 0; ientry < associativity; ientry++){
      count = write(fd, &invalidEntry, sizeof(HCEntry));
      if(count != sizeof(HCEntry)){
	perror("ERROR Creating HCCache file");
	exit(-1);
      }
    }
  }
}

/*
 *------------------------------------------------------------------
 *
 * HCDisk_Close --
 *
 *          Unmap the file.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
void HCDisk_Close(HCDisk *d)
{
  munmap((char *)d->mmappedArray, d->nbuckets * d->entriesPerBucket);
  d->mmappedArray = NULL;
}

/*
 *------------------------------------------------------------------
 *
 * HCDIsk_informLocalCopy --
 *
 *          Update data structures to reflect the fact a copy is
 *          cached locally.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
void HCDisk_informLocalCopy(HCDisk *d, URLKey key)
{
  HCEntry *bp;
  HCEntry new, dummyRet, old;
  int foundOld;

  assert(d);

  if(!URLKey_Compare(&key, &INVALID_URL_KEY)){
    return;
  }

  bp = bucket(d, key);
  foundOld = findKey(d, bp, key, &old);
  if(foundOld){
    deleteKey(d, bp, &old, &dummyRet);
  }
  HCEntry_Init(&new, key, (struct sockaddr_in *) getMyAddr());
  insertKey(d, bp, &new);
  if(!CACHE_RECENT_WRITES){
    if(madvise(pageAddr((caddr_t)bp), pageSize, MADV_DONTNEED)){
      perror("Madvise DONTNEED failed\n");
    }
  }
}



/*
 *------------------------------------------------------------------
 *
 * HCDisk_InvalLocalCopy --
 *
 *          Update local data structures to reflect the fact
 *          that we no longer have a copy locally.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
void HCDisk_invalLocalCopy(HCDisk *d, URLKey key)
{
  HCEntry *bp, myent, dummyRet;

  if(!URLKey_Compare(&key, &INVALID_URL_KEY)){
    return;
  }

  HCEntry_Init(&myent, key, (struct sockaddr_in *) getMyAddr());
  bp = bucket(d, key);
  deleteKey(d, bp, &myent, &dummyRet);
  if(madvise(pageAddr((caddr_t)bp), pageSize, MADV_DONTNEED)){
    perror("Madvise DONTNEED failed\n");
  }
}

/*
 *------------------------------------------------------------------
 *
 * HCPrefetch --
 *
 *          description.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
void 
HCDisk_prefetch(HCDisk *d, HCUpdate *uArray, int nupdates)
{
  HCEntry *bp;
  int ii;
  if(DO_PREFETCH) {
    for(ii = 0; ii < nupdates; ii++){
      if(uArray[ii].action == HC_InvalToParent ||
	 uArray[ii].action == HC_InvalToChild ||
	 uArray[ii].action == HC_InformToParent ||
	 uArray[ii].action == HC_InformToChild){
	bp = bucket(d, uArray[ii].entry.key);
	if(madvise(pageAddr((caddr_t)bp), pageSize, MADV_WILLNEED)){
	  perror("Madvise WILLNEED failed\n");
	}
      }
    }
  }
}


/*
 *------------------------------------------------------------------
 *
 * HCDisk_netInvalRecord --
 *
 *          The network told us that <key, invalNode> is
 *          no longer a valid mapping. If we knew of a 
 *          different copy than the one that was invalidated
 *          then report that back by returning 1 and putting
 *          a copy of the survivor in survivor.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
int
HCDisk_netInvalRecord(HCDisk *d, HCEntry *entry, HCEntry *survivor)
{
  HCEntry *bp;
  int remainingCopy;

  if(!HCE_VALID(entry)) {
    survivor->key.key = 0;
    return 0;
  }

  bp = bucket(d, entry->key);
  remainingCopy = deleteKey(d, bp, entry, survivor);
  if(madvise(pageAddr((caddr_t)bp), pageSize, MADV_DONTNEED)){
    perror("Madvise DONTNEED failed\n");
  }
  return remainingCopy;
}

/*
 *------------------------------------------------------------------
 *
 * HCDisk_updateIfCloser --
 *
 *          The network told us that <key, newnode> is a valid
 *          mapping. Install it if it is better than the one
 *          we have.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
int 
HCDisk_updateIfCloser(HCDisk *d, HCEntry *new)
{
  URLKey key = new->key;
  HCEntry old, *bp, dummyRet;
  int ret = 0;
  int foundOld;

  if(!HCE_VALID(new)){
    return 0;
  }

  bp = bucket(d, key);
  foundOld = findKey(d, bp, key, &old);
  if (!foundOld) {
    insertKey(d, bp, new);
    ret = 1;
  }
  else if(HCHier_compareDistanceFromMe(new->ipaddr, old.ipaddr) < 0){
    deleteKey(d, bp, &old, &dummyRet);
    insertKey(d, bp, new);
    ret = 1;
  }
  if(!CACHE_RECENT_WRITES){
    if(madvise(pageAddr((caddr_t)bp), pageSize, MADV_DONTNEED)){
      perror("Madvise DONTNEED failed\n");
    }
  }
  return ret;
}

/*
 *------------------------------------------------------------------
 *
 * HCDisk_findNearest --
 *
 *          Find the record for the key if one exists. Return 1
 *          if matching record found and a copy in *match. 
 *          Return 0 if no match found.
 *
 *          Note: we can't just return a pointer to match
 *          because its location in the bucket can change
 *          at any time.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *---------------------------------------------3.24.1998----------------
 */
int
HCDisk_findNearest(HCDisk *d, URLKey key, HCEntry *match)
{
  HCEntry *bp;
  int found;

  assert(match != 0);
  if(!URLKey_Compare(&key, &INVALID_URL_KEY)){
    return 0;
  }

  bp = bucket(d, key);
  found = findKey(d, bp, key, match);
  if(!CACHE_RECENT_READS){
    if(madvise(pageAddr((caddr_t)bp), pageSize, MADV_DONTNEED)){
      perror("Madvise DONTNEED failed\n");
    }
  }
  return found;
}


/*
 *------------------------------------------------------------------
 *
 * bucket --
 *
 *          Return pointer to first entry in bucket for key.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
static HCEntry *
bucket(HCDisk *d, URLKey key)
{
  int b;

  assert(d);
  assert(d->mmappedArray);

  b = key.key % d->nbuckets;
  return &(d->mmappedArray[b * d->entriesPerBucket]);
}

/*
 *------------------------------------------------------------------
 *
 * pageAddr --
 *
 *          Return the address of the first byte of the
 *          page that holds an address.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
static caddr_t
pageAddr(caddr_t e)
{
  caddr_t ret;
  ret = e - ((unsigned)e % pageSize);
  return ret;
}

/*
 *------------------------------------------------------------------
 *
 * deleteKey --
 *
 *          Delete the record associated with the specified <key, node>
 *          pair.
 *          
 *          If there is a record for the key that survives the
 *          delete (e.g. the node doesn't match) return it
 *          in the HCEntry *surviver field and return 1. If
 *          no record for the key survives the delete, return 0.
 *          
 *          Note: we can't just return a pointer to survivor
 *          because its location in the bucket can change
 *          at any time.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *---------------------------------------------------5.5.1998---------
 */
static int
deleteKey(HCDisk *d, HCEntry *bucketp, HCEntry *entry, HCEntry *survivor)
{
  int ii;

  assert(survivor != NULL);
  assert(entry != survivor); 
  assert(HCE_VALID(entry));

  for(ii = 0; ii < d->entriesPerBucket; ii++){
    if(!URLKey_Compare(&bucketp[ii].key, &entry->key)){
      if(!HCEntry_Compare(entry, &bucketp[ii])){
	HCE_INVALIDATE(&bucketp[ii]);
	sanityCheckMatch(d, bucketp, entry, 0, 0);
	return 0;
      }
      else{
	/*
	 * Matched key but not NodeId.
	 */
	*survivor = bucketp[ii];
	sanityCheckMatch(d, bucketp, entry, 0, 1);
	sanityCheckMatch(d, bucketp, survivor, 1, 1);
	return 1;
      }
    }
  }
  /*
   * No match key
   */
  return 0;
}


/*
 *------------------------------------------------------------------
 *
 * insertKey --
 *
 *          Insert an entry in a given bucket in hash table.
 *          The table is n-way associative. The new entry
 *          is always inserted into entry 0, and the remaining
 *          entries are shifted up. If there is an empty
 *          entry, we don't have to shift any higher entries.
 *          If all entries are full, entry d->entriesPerBucket - 1 
 *          (the last entry) is lost.
 *
 * Arguments:
 *      None.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *-------------------------------------------------5.5.1998----------
 */
static void insertKey(HCDisk *d, HCEntry *bucketp, HCEntry *entry)
{
  int from, to;

  /*
   * Caller responsible for deleting old entry before inserting
   * a new one.
   */
  sanityCheckMatch(d, bucketp, entry, 0, 0);

  /*
   * First, see if there are any empty entries. All entries
   * below the first empty one get shifted up to make
   * room in entry 0. The last entry by definition is
   * emtpy if all earlier ones are empty (you are always
   * allowed to shift over the last entry.)
   */
  for(to = 0; to < d->entriesPerBucket - 1; to++){
    if(!HCE_VALID(&bucketp[to])){
      break;
    }
  }
  assert(to <= d->entriesPerBucket -1 );
  assert(!HCE_VALID(&bucketp[to]) || (to == d->entriesPerBucket - 1));

  /*
   * Shift entry <from> up to entry <from+1> to make entry 0 empty.
   * The last bucket we shift is entry (from==entriesPerBucket-2)
   * which shifts into the last bucket (entriesPerBucket-1).
   * Don't go any farther, or we clobber the next bucket.
   */
  for(; to > 0; to--){
    from = to - 1;
    assert(from >= 0);
    bucketp[to] = bucketp[from];
  }
  bucketp[0] = *entry;
  sanityCheckMatch(d, bucketp, entry, 1, 1);
}


/*
 *------------------------------------------------------------------
 *
 * sanityCheckMatch --
 *
 *      Check to see that the expected number of entries in bucket
 *      match specified key.
 *
 * Arguments:
 *      None.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *---------------------------------------------------5.6.1998-------
 */
static void
sanityCheckMatch(HCDisk *d, HCEntry *bucketp, HCEntry *entry, 
		 int expectFullMatch, int expectKeyMatch)
{
  int ii;
  int fullMatch = 0, keyMatch = 0;

  for(ii = 0; ii < d->entriesPerBucket; ii++){
    if(!HCEntry_Compare(&bucketp[ii], entry)){
      fullMatch++;
    }
    if(!URLKey_Compare(&bucketp[ii].key, &entry->key)){
      keyMatch++;
    }
  }
  assert(fullMatch == expectFullMatch);
  assert(keyMatch == expectKeyMatch);
}

/*
 *------------------------------------------------------------------
 *
 * findKey --
 *
 *          Find an entry that matches the key. Return
 *          true if found and put the result in foundRet.
 *          Since this is a read, 
 *          update the LRU list by removing key and
 *          then re-inserting key.
 *         
 *          Note: we can't just return a pointer to copy
 *          because its location in the bucket can change
 *          at any time.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
static int
findKey(HCDisk *d, HCEntry *bucketp, URLKey key, HCEntry *copy)
{
  int ii;

  assert(copy != NULL);
  assert(copy < bucketp || copy > &bucketp[d->entriesPerBucket-1]);
  assert(URLKey_Compare(&key, &INVALID_URL_KEY));
  for(ii = 0; ii < d->entriesPerBucket; ii++){
    if(!URLKey_Compare(&bucketp[ii].key, &key)){
      *copy = bucketp[ii];
      assert(!URLKey_Compare(&copy->key, &key));
      if(ii != 0){
	/*
	 * Move to front of LRU list by deleting entry and inserting
	 */
	HCE_INVALIDATE(&bucketp[ii]);
	insertKey(d, bucketp, copy);
      }
      return 1;
    }
  }
  HCE_INVALIDATE(copy);
  return 0;
}

#ifdef DOTEST
/*
 *------------------------------------------------------------------
 *
 * HCDisk_SelfTest --
 *
 *          description.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
void
HCDisk_selfTest()
{
  HCDisk d;
  char urlString[1024];
  URLKey url, oldUrl, olderUrl;
  /* Nodes */
  struct sockaddr_in n0, n1, n2, n3, n4, n5, n6;
  HCEntry place, newEntry;
  int size, assoc, ii;
  int maxSize, maxAssoc, maxi;
  const static int doBigTest = 0;
  struct sockaddr_in *s;
  int found;
  char name[128];
  char testPath[256];

  sprintf(testPath, "DiskSelfTestFile-%d-%s", (int)getpid(), 
	  inet_ntoa(getMyAddr()->sin_addr));

  if(doBigTest){
    maxSize = 10000;
    maxAssoc = 8;
    maxi = 100000;
  }
  else{
    maxSize = 1000;
    maxAssoc = 5;
    maxi = 100;
  }


  printf("HCDisk_selfTest..."); fflush(stdout);

  timeMadvise();

  n0.sin_addr.s_addr = 0x80536F00; n0.sin_port = 0;
  n1.sin_addr.s_addr = 0x81536F01; n1.sin_port = 1;
  n2.sin_addr.s_addr = 0x82536F02; n2.sin_port = 2;
  n3.sin_addr.s_addr = 0x83536F03; n3.sin_port = 3;
  n4.sin_addr.s_addr = 0x84536F04; n4.sin_port = 4;
  n5 = *getMyAddr();
  n6.sin_addr.s_addr = 0x86536F06; n6.sin_port = 6;
  s = &n5; /* I am node 5 */

  /* Put hopcount in netdb as initial distance metric. */
  /* Need to store hostname in a buffer so that callee can use inet_ntoa */
  strcpy(name, inet_ntoa(n0.sin_addr));
  netdbAdd(n0.sin_addr, name);
  netdbHandlePingReply(&n0, 3, 0);
  assert(netdbHostHops(name) == 3);

  strcpy(name, inet_ntoa(n1.sin_addr));
  netdbAdd(n1.sin_addr, name);
  netdbHandlePingReply(&n1, 2, 0);
  assert(netdbHostHops(name) == 2);

  strcpy(name, inet_ntoa(n2.sin_addr));
  netdbAdd(n2.sin_addr, name);
  netdbHandlePingReply(&n2, 1, 0);
  assert(netdbHostHops(name) == 1);

  strcpy(name, inet_ntoa(n3.sin_addr));
  netdbAdd(n3.sin_addr, name);
  netdbHandlePingReply(&n3, 2, 0);
  assert(netdbHostHops(name) == 2);

  strcpy(name, inet_ntoa(n4.sin_addr));
  netdbAdd(n4.sin_addr, name);
  netdbHandlePingReply(&n4, 2, 0);
  assert(netdbHostHops(name) == 2);

  strcpy(name, inet_ntoa(n5.sin_addr));
  netdbAdd(n5.sin_addr, name);
  netdbHandlePingReply(&n5, 0, 0);
  /*  assert(netdbHostHops(name) == 0);*/


  strcpy(name, inet_ntoa(n6.sin_addr));
  netdbAdd(n6.sin_addr, name);
  netdbHandlePingReply(&n6, 1, 0);
  assert(netdbHostHops(name) == 1);

  /* * This stuff will only work if netdb is working
   * and compareDistanceFromMe is working */
assert(HCHier_compareDistanceFromMe(n0.sin_addr, n1.sin_addr) > 0);
assert(HCHier_compareDistanceFromMe(n1.sin_addr, n0.sin_addr) < 0);
assert(HCHier_compareDistanceFromMe(n3.sin_addr, n4.sin_addr) == 0);
assert(HCHier_compareDistanceFromMe(n6.sin_addr, n1.sin_addr) < 0);

  for(size = 8 * sizeof(HCEntry); size < maxSize; size *= 10){ 
    for(assoc = 1; assoc < maxAssoc; assoc+=3){
      printf("."); fflush(stdout);
      HCDisk_CreateFile(testPath, size, assoc);
      HCDisk_Init(&d, testPath);

      url = INVALID_URL_KEY;
      oldUrl = INVALID_URL_KEY;
      olderUrl = INVALID_URL_KEY;

      for(ii = 0; ii < maxi; ii++){
	olderUrl = oldUrl;
	oldUrl = url;
	sprintf(urlString, "http://wlksjdf%ddlkfjskdf", ii);
	URLKey_Init(&url, urlString);

	/*
	 * Check miss 
	 */
	found = HCDisk_findNearest(&d, url, &place);
	assert(!found);

	/*
	 * Write then read
	 */
	HCEntry_Init(&newEntry, url, &n1);
	assert(HCDisk_updateIfCloser(&d, &newEntry));
	found = HCDisk_findNearest(&d, url, &place);
	assert(found);
	assert(place.ipaddr.s_addr == n1.sin_addr.s_addr
	       && place.port == n1.sin_port);

	/*
	 * check net overwrite with closer
	 */
	HCEntry_Init(&newEntry, url, &n6);
	assert(HCDisk_updateIfCloser(&d, &newEntry));
	found = HCDisk_findNearest(&d, url, &place);
	assert(found);
	assert(place.ipaddr.s_addr == n6.sin_addr.s_addr
	       && place.port == n6.sin_port);

	/*
	 * Check net no overwrite with farther
	 */
	HCEntry_Init(&newEntry, url, &n3);
	assert(!HCDisk_updateIfCloser(&d, &newEntry));
	found = HCDisk_findNearest(&d, url, &place);
	assert(found);
	assert(place.ipaddr.s_addr != n3.sin_addr.s_addr
	       && place.port != n3.sin_port);
	assert(place.ipaddr.s_addr == n6.sin_addr.s_addr
	       && place.port == n6.sin_port);

	/*
	 * Check net invalidate -- different copy
	 */
	HCEntry_Init(&newEntry, url, &n1);
	found = HCDisk_netInvalRecord(&d, &newEntry, &place);
	assert(found);
	assert(place.ipaddr.s_addr == n6.sin_addr.s_addr
	       && place.port == n6.sin_port);
	found = HCDisk_findNearest(&d, url, &place);
	assert(found);
	assert(place.ipaddr.s_addr == n6.sin_addr.s_addr
	       && place.port == n6.sin_port);



	/*
	 * Check net invalidate -- of the one we've got
	 */
	HCEntry_Init(&newEntry, url, &n6);
	found = HCDisk_netInvalRecord(&d, &newEntry, &place);
	assert(!found);
	found = HCDisk_findNearest(&d, url, &place);
	assert(!found);

	/*
	 * Check local update
	 */
	HCDisk_informLocalCopy(&d, url);
 	found = HCDisk_findNearest(&d, url, &place);
	assert(found);
	assert(place.ipaddr.s_addr == n5.sin_addr.s_addr
	       && place.port == n5.sin_port);

	/*
	 * Try to insert multiple local copies. Should be filtered.
	 */
	HCDisk_informLocalCopy(&d, url);
	HCDisk_informLocalCopy(&d, url);
	HCDisk_informLocalCopy(&d, url);

	/*
	 * Check local invalidate
	 */
	HCDisk_invalLocalCopy(&d, url);
	found = HCDisk_findNearest(&d, url, &place);
	assert(!found);


	/* 
	 * Leave local copy when done
	 */
	HCDisk_informLocalCopy(&d, url);
	found = HCDisk_findNearest(&d, url, &place);
	assert(found);
	assert(place.ipaddr.s_addr == n5.sin_addr.s_addr
	       && place.port == n5.sin_port);

	/*
	 * check associativity by reading from past passes
	 */
	if(assoc > 1){
	  if(URLKey_Compare(&oldUrl, &INVALID_URL_KEY)){
	    found = HCDisk_findNearest(&d, oldUrl, &place);
	    assert(found);
	    assert(place.ipaddr.s_addr == n5.sin_addr.s_addr
		   && place.port == n5.sin_port);
	  }
	}
	if(assoc > 2){
	  if(URLKey_Compare(&olderUrl, &INVALID_URL_KEY)){
	    found = HCDisk_findNearest(&d, olderUrl, &place);
	    assert(found);
	    assert(place.ipaddr.s_addr == n5.sin_addr.s_addr
		   && place.port == n5.sin_port);
	  }
	}

      }  
      HCDisk_Close(&d);
      remove(testPath);
    }
  }
  printf("..OK\n");
}


static void
timeMadvise()
{
  struct timeval start, end;
  char *data, *p;
  int ii;
  float time;
  static const int count = 10000;

  if((pageSize = sysconf(_SC_PAGESIZE)) < 0){
    perror("Cannot determine page size\n");
    exit(-1);
  }

  data = (char *)malloc(16382);
  p = pageAddr(data);

  assert(!gettimeofday(&start, NULL));
  for(ii = 0; ii < count; ii++){
    madvise(p, pageSize, MADV_DONTNEED);
  }
  assert(!gettimeofday(&end, NULL));
  time =  end.tv_sec - start.tv_sec;
  time = time + (float)(end.tv_usec - start.tv_usec)/(float)(1000000.0);

  printf("Called Madvise %d times in %f seconds: %f per call\n",
	 count, time, time/(float)count);
  free(data);
}
#endif /* DOTEST */
