/*
 *------------------------------------------------------------------
 *
 * LESS Group 
 * Computer Sciences Department 
 * University of Texas at Austin
 *
 * HCHier ---
 *          Abstraction of the logical hierarchy of nodes.
 *
 * $Date: 1998/05/06 21:14:54 $ $Id: HCHier.c,v 1.18 1998/05/06 21:14:54 dahlin Exp $
 *
 * Copyright (c) 1997 by Mike Dahlin.
 * 
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for noncommercial use without fee and
 * without written agreement is hereby granted, provided that 
 * the above copyright notice appear in all copies of this software.
 * 
 *------------------------------------------------------------------
 */
#include <assert.h>

#define FRIEND_OF_HCNODELIST  1  /* so we can see .h file */
#define FRIEND_OF_HCPLAX 1
#include "squid.h"
#include "HCNodelist.h"
#include "HCHintCache.h"
static int doTestBarrier(int barrier);
static void writeTestBarrier(int barrier);
static int readTestBarrier(int barrier);
static void clearTestBarrier();
static void testFindPeer(peer *p);
static void testPlax(void);
static void testCreateStuff(char *tag);
static void testFind(void);
static void testFindStuff(char *tag, const struct sockaddr_in *expect);

/*
 * These must be an set of values each one greater than the
 * last to define the series of states for the network test.
 */
#define BARRIER_START 1000
#define BARRIER_READY 1001
#define BARRIER_JOIN 1002
#define BARRIER_JOIN2 1003
#define BARRIER_NODELIST 1004
#define BARRIER_CREATE_STUFF 1005
#define BARRIER_CREATE_STUFF2 1006
#define BARRIER_CREATE_STUFF3 1007
#define BARRIER_PLAX 1008
#define BARRIER_FIND 1009
#define BARRIER_LAST 1010
#define BARRIER_DONE 9999

int HCHier_selfTestDone = 0;


#define USE_DYNAMIC_HIERARCHY 1

List_Links HCHier_Nodes;

/*
 *------------------------------------------------------------------
 *
 * HCHier_Init --
 *
 *          Constructor. 
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *-------------------------------------------------4.16.1998-----------
 */
void
HCHier_Init(void)
{
  HCNLIter iter;
  struct sockaddr_in saddr;
  /*
   * This should be set in config file
   */
  static int bitsPerLevelXXX = 8;

  HCNodelist_Init();
  HCNodelist_LocalJoin();
  HCPlax_Init(bitsPerLevelXXX);
  for(HCNodeList_IterStart(&iter);
      HCNodeList_IterCheck(iter);
      HCNodeList_IterNext(&iter)){
    if(HCNodeList_IterGetAddr(iter, &saddr)){
      HCPlax_addNode(&saddr);
    }
  }
}

/*
 *------------------------------------------------------------------
 *
 * HCHier_Destroy --
 *
 *          description.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *-------------------------------------------------4.1.1998------------
 */
void
HCHier_Destroy(void)
{
  HCNodelist_LocalLeave();
  HCNodelist_Destroy();
  HCPlax_Destroy();
}

/*
 *------------------------------------------------------------------
 *
 * HCHier_compareDistanceFromMe --
 *
 *          Return <0 if node1 closer than node2, 0 if same
 *          distance, >0 if node2 closer than node1. If either
 *          node is missing, it must be far away. If both
 *          are missing, venture no opinion (0).
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
int
HCHier_compareDistanceFromMe(struct in_addr new, struct in_addr old)
{
    int nrtt, ortt;
    int nhops, ohops;
    char name[16];

    /* Compare by rtt if both have measured rtts */
    /* Must make copy of result of inet_ntoa before making
     * function call because netdb also uses that function */
    strcpy(name, inet_ntoa(new));
    nrtt = netdbHostRtt(name);
    strcpy(name, inet_ntoa(old));
    ortt = netdbHostRtt(name);
    if (nrtt && ortt)
	return(nrtt - ortt);

    /* Otherwise compare hopcount */
    strcpy(name, inet_ntoa(new));
    nhops = netdbHostHops(name);
    strcpy(name, inet_ntoa(old));
    ohops = netdbHostHops(name);
    return(nhops - ohops);
}

/*
 *------------------------------------------------------------------
 *
 * HCHier_newNode --
 *
 *          This gets called when we hear about a new node in the system.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *-------------------------------------------------4.1.1998-----------
 */
void
HCHier_newNode(struct sockaddr_in *sin, int hops, 
	       long long joinTimeNS)
{
  HCNodelist_NetJoin(sin, hops, joinTimeNS);
}

/*
 *------------------------------------------------------------------
 *
 * HCHier_delNode --
 *
 *          This gets called when we hear about a node leaving the system
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *-------------------------------------------------4.1.1998-----------
 */
void
HCHier_delNode(struct sockaddr_in *sin, int hops, 
	       long long joinTimeNS)
{
    HCNodelist_NetLeave(sin, hops, joinTimeNS);
}





/*
 *------------------------------------------------------------------
 *
 * HCHier_CheckIfParent --
 *
 *          See if the specified node is really my parent.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *---------------------------------------------------4.10.1998---------
 */
int 
HCHier_CheckIfParent (URLKey key, struct sockaddr_in *candidate)
{
  if(USE_DYNAMIC_HIERARCHY){
    return HCPlax_CheckIfParent(key, candidate);
  }
  else{
    /* 
     * For static algorithm, anyone that wants to be my parent can 
     * be.
     */
    return 1;
  }
}

/*
 *------------------------------------------------------------------
 *
 * HCHier_CheckIfChild --
 *
 *          See if the specified node is really my child.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *---------------------------------------------------4.10.1998---------
 */
int 
HCHier_CheckIfChild (struct sockaddr_in *candidate)
{
  if(USE_DYNAMIC_HIERARCHY){
    return HCPlax_CheckIfChild(candidate);
  }
  else{
    /* 
     * For static algorithm, anyone that wants to be my child can 
     * be.
     */
    return 1;
  }
}

/*
 *------------------------------------------------------------------
 *
 * HCHier_AddChild --
 *
 *          This node has claimed to by my child, so add it
 *          to the dynamic hierarchy.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------4.10.1998--------
 */
void 
HCHier_AddChild (struct sockaddr_in *child)
{
  if(USE_DYNAMIC_HIERARCHY){
    HCPlax_AddChild(child);
  }
  else{
    /*
     * For static algorithm, I don't track children.
     */
    return;
  }
}


/*
 *------------------------------------------------------------------
 *
 * HCHier_RemoveChild --
 *
 *          This node denies paternity. Disinherit it.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *---------------------------------------------------4.10.1998----------
 */
void HCHier_RemoveChild (struct sockaddr_in *child)
{
  if(USE_DYNAMIC_HIERARCHY){
    HCPlax_RemoveChild(child);
  }
  else{
    /*
     * For static algorithm, I don't track children.
     */
    return;
  }
}



/*
 *------------------------------------------------------------------
 *
 * HCHier_ChildQIndex --
 *
 *          Return the index into the array of queues of 
 *          children for children that are my children 
 *          for this object. (See HCPlax for how
 *          children are grouped into levels.)
 *
 *          Also, if I have no children for this object,
 *          set amILeaf to true.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------4.10.1998-------------
 */
int 
HCHier_ChildQIndex(URLKey *key, int iAmRoot, int *amILeaf)
{
  assert(amILeaf != NULL); 
  if(USE_DYNAMIC_HIERARCHY){
    return HCPlax_ChildQIndex(key, iAmRoot, amILeaf);
  }
  else{
    /*
     * For static algorithm, we have one set of children and
     * thus one outgoing children buffer with index 0.
     */
    *amILeaf = 0;
    return 0;
  }
}

/*
 *------------------------------------------------------------------
 *
 * HCHier_NChildrenQs --
 *
 *          Return the number of outgoing message queues
 *          to different classes of children. Each
 *          "class" corresponds to children for which 
 *          I am a parent. (See HCPlax for how
 *          children are grouped into levels.)
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------4.10.1998------------
 */
int HCHier_NChildrenQs(void)
{
  if(USE_DYNAMIC_HIERARCHY){
    return HCPlax_NChildrenQs();
  }
  else{
    /*
     * For static algorithm, we have one set of children and
     * thus one outgoing children buffer with index 0.
     */
    return 1;
  }
}



/*
 *------------------------------------------------------------------
 *
 * HCHier_GetChildAddrs --
 *
 *        Return the number of children for the specified message
 *        queue, allocate a buffer for that many addresses,
 *        and fill in that bufer with the children's addresses.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      See above.
 *
 * Side effects:
 *      Allocates a vector for the addresses. Caller must free
 *      this vector when done.
 *
 *      Note: the vector may be out of date by the time ther caller
 *      uses it. This is inherent problem -- even if caller sent
 *      all messages instantaneously, by the time they arrive
 *      at destination, the child could have left the system
 *      or a new child could have joined.
 *
 *------------------------------------------------4.16.1998------------
 */
int 
HCHier_GetChildAddrs(int qIndex, struct sockaddr_in **retAddrAP)
{
  peer *p;
  int count = 0, nchildren = 0;

  assert(retAddrAP);
  assert(qIndex < HCHier_NChildrenQs());

  if(USE_DYNAMIC_HIERARCHY){
    return HCPlax_GetChildAddrs(qIndex, retAddrAP);
  }
  else{
    assert(qIndex == 0);
    for (p = getFirstPeer(); p; p = getNextPeer(p)) {
      nchildren++;
    }
    *retAddrAP = (struct sockaddr_in *)xmalloc(nchildren
					     * sizeof(struct sockaddr_in));
    for (p = getFirstPeer(); p; p = getNextPeer(p)) {
      (*retAddrAP)[count] = p->in_addr;
      count++;
    }
    /*
     * Squid promises that we run atomically
     */
    assert(count == nchildren);
    return nchildren;
  }
}

/*
 *------------------------------------------------------------------
 *
 * HCHier_NParentQs --
 *
 *          How many different message queues are they for different
 *          parents?
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------4.10.1998---------
 */
int HCHier_NParentQs(void)
{
  if(USE_DYNAMIC_HIERARCHY){
    return HCPlax_NParentQs();
  }
  else{
    /*
     * For static algorithm, we have one parent
     * thus one outgoing parent buffer with index 0.
     */
    return 1;
  }
}


/*
 *------------------------------------------------------------------
 *
 * HCHier_ParentQIndex --
 *
 *          Return the index of the queue I should use to talk
 *          to the parent for the specified object (set amIRoot
 *          if I am the root for the object.)
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *---------------------------------------------------4.10.1998---------
 */
int HCHier_ParentQIndex(URLKey *key, int *amIRoot)
{
  assert(amIRoot != NULL);
  if(USE_DYNAMIC_HIERARCHY){
    return HCPlax_ParentQIndex(key, amIRoot);
  }
  else{
    /*
     * For static algorithm, we have one parent
     * thus one outgoing parent buffer with index 0.
     */
    *amIRoot = 0;
    return 0;
  }
}


/*
 *------------------------------------------------------------------
 *
 * HCHier_GetParentAddr --
 *
 *          Return the address of the node that is curerntly
 *          the parent for the specified message queue.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      Put answer in *retAddr and return nonzero on error.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------4.10.1998-------
 */
int HCHier_GetParentAddr(int index, struct sockaddr_in *retAddr)
{
  peer *p;
  if(USE_DYNAMIC_HIERARCHY){
    return HCPlax_GetParentAddr(index, retAddr);
  }
  else{
    for (p = getFirstPeer(); p; p = getNextPeer(p)) {
      if(p->type == PEER_PARENT){
	*retAddr = p->in_addr;
	return 0;
      }
    }
    p = getFirstPeer();
    if(p){
      *retAddr = p->in_addr;
      return 0;
    }
    return -1;
  }
}



/*
 *------------------------------------------------------------------
 *
 * HCHier_selfTest --
 *
 *          description.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *---------------------------------------------------4.24.1998-----------
 */
void
HCHier_selfTest()
{
  HCNodelist_SelfTest();
  HCPlax_SelfTest();

  /*
   * Note:
   * HCHier_testNetwork() gets called repeatedly elsewhere (from HCTimer)
   * not from normal self test routines because it needs the
   * network to be working, so it needs to "yield" to main event
   * loop.
   */

#ifdef DOTEST_NOTYET
  char *testHier = "testHierarchyFile";
  char *testBadPath = "/nonexistantFileName";
  char *testBadFile = "testHierarchyFile.Bad";
  char *testLoopFile = "testHierarchyFile.loop";
  char *testBadChildrenFile = "testHierarchyFile.badChildren";

  HCHier h;
  NodeId myId, expect;
  struct sockaddr *s;
  NodeId n, n1, n2;

  printf("HCHier_selfTest().");

  NodeId_Init(&myId, 0x80536F03, 2);
  s = NodeId_copyAsSockAddr(&myId);

  assert(HCHier_Init(&h, testBadPath, s) == -1);
  assert(HCHier_Init(&h, testHier, s) == 0);
  assert(HCHier_nChildren(&h) == 3);
  assert(HCHier_nParents(&h) == 1);

  /*
   * Test children
   */
  n = HCHier_getChild(&h, 0);
  NodeId_Init(&expect, 0x80536F03, 5);
  assert(!NodeId_Compare(&n, &expect));
  n = HCHier_getChild(&h, 1);
  NodeId_Init(&expect, 0x80536F03, 4);
  assert(!NodeId_Compare(&n, &expect));
  n = HCHier_getChild(&h, 2);
  NodeId_Init(&expect, 0x80536F03, 3);
  assert(!NodeId_Compare(&n, &expect));

  /*
   * Test parents
   */
  n = HCHier_getParent(&h, 0);
  NodeId_Init(&expect, 0x80536F03, 1);
  assert(!NodeId_Compare(&n, &expect));


  /* 
   * Test distance 
   */
  NodeId_Init(&n1, 0x80536F03, 0);
  NodeId_Init(&n2, 0x80536F03, 1);
  assert(HCHier_compareDistanceFromMe(&h, n1, n2) > 0);
  assert(HCHier_compareDistanceFromMe(&h, n2, n1) < 0);

  NodeId_Init(&n1, 0x80536F03, 5);
  NodeId_Init(&n2, 0x80536F03, 6);
  assert(HCHier_compareDistanceFromMe(&h, n1, n2) < 0);
  assert(HCHier_compareDistanceFromMe(&h, n2, n1) > 0);

  NodeId_Init(&n1, 0x80536F03, 3);
  NodeId_Init(&n2, 0x80536F03, 4);
  assert(HCHier_compareDistanceFromMe(&h, n1, n2) == 0);
  assert(HCHier_compareDistanceFromMe(&h, n2, n1) == 0);

  /*
   * My node id 
   */
  n = HCHier_myNodeId(&h);
  assert(!NodeId_Compare(&n, &myId));
  

  /*
   * Subtree
   */
  NodeId_Init(&n1, 0x80536F03, 1);
  assert(!HCHier_inMySubtree(&h, n1));
  NodeId_Init(&n1, 0x80536F03, 2);
  assert(!HCHier_inMySubtree(&h, n1));
  NodeId_Init(&n1, 0x80536F03, 4);
  assert(HCHier_inMySubtree(&h, n1));
  NodeId_Init(&n1, 0x80536F03, 6);
  assert(HCHier_inMySubtree(&h, n1));

  HCHier_Destroy(&h);
  printf("\nHCHier -- trying bad input files. Expect warnings\n");
  printf("\nHCHier -- bad input file (bad format): ");
  assert(HCHier_Init(&h, testBadFile, s) == -1);
  printf("\nHCHier -- bad input file (loop): ");
  assert(HCHier_Init(&h, testLoopFile, s) == -1);
  printf("\nHCHier -- bad input file (missing children): ");
  assert(HCHier_Init(&h, testBadChildrenFile, s) == -1);

  printf("..OK\n");
#endif /* DOTEST_NOTYET */
}



/*
 *------------------------------------------------------------------
 *
 * HCHier_testNetwork --
 *
 *          Test that we can create a hierarchy with another 
 *          node on the network.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *-------------------------------------------------4.19.1998---------
 */
void
HCHier_testNetwork(void)
{
  peer *p;
  int ipeer = 0;
  static int phase = BARRIER_START;
  char tag[256];

  assert(USE_DYNAMIC_HIERARCHY);

  if(phase == BARRIER_START){
    printf("XXX Testing hint cache network stuff \n");
    
    for (p = getFirstPeer(); p; p = getNextPeer(p)) {
      printf("Peer %d: %s %s\n", ipeer, p->host, inet_ntoa(p->in_addr.sin_addr));
      ipeer++;
    }
    
    printf("XXX My hostname %s address %s\n", getMyHostname(), 
	   inet_ntoa(getMyAddr()->sin_addr));

    clearTestBarrier();
    phase = BARRIER_READY;
    return;
  }

  if(phase == BARRIER_READY){
    if(!doTestBarrier(BARRIER_READY)){
      return;
    }
    phase = BARRIER_JOIN;
    return;
  }

  if(phase == BARRIER_JOIN){
    if(!doTestBarrier(BARRIER_JOIN)){
      return;
    }
    phase = BARRIER_JOIN2;
    return;
  }

  if(phase == BARRIER_JOIN2){
    if(!doTestBarrier(BARRIER_JOIN2)){
      return;
    }
    phase = BARRIER_NODELIST;
    return;
  }

  if(phase == BARRIER_NODELIST){
    for(p = getFirstPeer(); p; p = getNextPeer(p)){
      testFindPeer(p);
    }
    if(!doTestBarrier(BARRIER_NODELIST)){
      return;
    }
    phase = BARRIER_CREATE_STUFF;
    return;
  }

  if(phase == BARRIER_CREATE_STUFF){
    sprintf(tag, "%s", inet_ntoa(getMyAddr()->sin_addr));
    testCreateStuff(tag);
    testCreateStuff("shared");
    if(!doTestBarrier(BARRIER_CREATE_STUFF)){
      return;
    }
    phase = BARRIER_CREATE_STUFF2;
  }

  if(phase == BARRIER_CREATE_STUFF2){
    /*
     * Just making sure that data from last phase makes it around
     * the system
     */
    if(!doTestBarrier(BARRIER_CREATE_STUFF2)){
      return;
    }
    phase = BARRIER_CREATE_STUFF3;
  }

  if(phase == BARRIER_CREATE_STUFF3){
    /*
     * Just making sure that data from last phase makes it around
     * the system
     */
    static int passes = 0;
    if(passes < 2){
      passes++;
      return;
    }
    if(!doTestBarrier(BARRIER_CREATE_STUFF3)){
      return;
    }
    phase = BARRIER_PLAX;
  }


  if(phase == BARRIER_PLAX){
    testPlax();
    if(!doTestBarrier(BARRIER_PLAX)){
      return;
    }
    phase = BARRIER_FIND;
    return;
  }

  if(phase == BARRIER_FIND){
    testFind();
    if(!doTestBarrier(BARRIER_FIND)){
      return;
    }
    phase = BARRIER_LAST;
    return;
  }

  if(phase == BARRIER_LAST){
    if(!doTestBarrier(BARRIER_LAST)){
      return;
    }
    phase = BARRIER_DONE;
    return;
  }

  if(phase == BARRIER_DONE){
    if(!doTestBarrier(BARRIER_DONE)){
      return;
    }
    HCHier_selfTestDone = 1;
    printf("HCHier multi-node selfTest Done\n");
    return;
  }

  assert(0);
}

/*
 *------------------------------------------------------------------
 *
 * doTestBarrier --
 *
 *          Write the flag to the barrier and then read
 *          the flag from the barrier files written
 *          by the other nodes.
 *
 *          If all of the other barrier files show
 *          the expected flag or later, return true
 *          else return false.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *---------------------------------------------------4.19.1998--------
 */
static int
doTestBarrier(int barrier)
{
  static int lastBarrier = 0;

  if(barrier != lastBarrier){
    printf("XXX To barrier %d\n", barrier);
    lastBarrier = barrier;
  }
  writeTestBarrier(barrier);
  return readTestBarrier(barrier);
}

/*
 *------------------------------------------------------------------
 *
 * writeBarrier --
 *
 *          Write the specified flag to my barrier file
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *---------------------------------------------------------4.19.1998--
 */
static void
writeTestBarrier(int barrier)
{
  int fd;
  char name[128];
  char data[128];
  int len, got;

  sprintf(name, "%s-%s", "tmp.barrier", getMyHostname());
  fd = open(name, O_WRONLY|O_CREAT|O_TRUNC, 0666);
  if(fd < 0){
    perror("HCHier: open barrier");
    assert(0);
  }
  sprintf(data, "FLAG %d\n", barrier);
  len = strlen(data);
  got = write(fd, data, len);
  assert(got == len);
  close(fd);
  return;
}

/*
 *------------------------------------------------------------------
 *
 * readBarrier --
 *
 *          Read the neighbors' barrier files. If all neighbors
 *          show a flag with the value expected or one greater, 
 *          return true. (One greater means -- we all reached
 *          this barrier, but the other node is now waiting at
 *          the next one.)
 *          Otherwise, return false.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------4.19.1998----
 */
static int
readTestBarrier(int barrier)
{
  int fd;
  char name[128];
  char data[128];
  int got;
  peer *p;
  int bval;

  for(p = getFirstPeer(); p; p = getNextPeer(p)){
    sprintf(name, "%s-%s", "tmp.barrier", p->host);
    fd = open(name, O_RDONLY);
    if(fd < 0){    /* other may not have gotten to barrier yet */
      return 0;
    }
    got = read(fd, data, 128);
    if(got <= 0){
      close(fd);
      return 0;
    }
    data[got]= '\0';
    if(sscanf(data, "FLAG %d", &bval) != 1){
      close(fd);
      return 0;
    }
    if((bval != barrier) && (bval != barrier + 1) 
       && (barrier != BARRIER_LAST || bval != BARRIER_DONE)){
      close(fd);
      return 0;
    }
    close(fd);
  }
  return 1;
}


/*
 *------------------------------------------------------------------
 *
 * clearTestBarrier --
 *
 *          Wipe out all test barriers. The first time I go through,
 *          I wipe out all barriers. This is OK. Even if others
 *          have written BARRIER_START to them, they will re-write
 *          the next time they come back.
 *
 * Arguments:
 *      None.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------4.24.1998----------
 */
static void
clearTestBarrier()
{
  int fd;
  char name[128];
  peer *p;
  
  for(p = getFirstPeer(); p; p = getNextPeer(p)){
    sprintf(name, "%s-%s", "tmp.barrier", p->host);
    fd = open(name, O_WRONLY|O_TRUNC|O_CREAT);
    if(fd >= 0){
      close(fd);
    }
  }
}

/*
 *------------------------------------------------------------------
 *
 * testFindPeer --
 *
 *          Test routine. Find the specified peer as a node
 *          on the nodelist or assert(0);
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
static void
testFindPeer(peer *p)
{
  HCNLIter i;
  struct sockaddr_in sin;

  for(HCNodeList_IterStart(&i);
      HCNodeList_IterCheck(i);
      HCNodeList_IterNext(&i)){
    if(HCNodeList_IterGetAddr(i, &sin)){
      printf("XXX testFindPeer() node found on node list %s\n",
	     inet_ntoa(sin.sin_addr));
      if(p->in_addr.sin_addr.s_addr == sin.sin_addr.s_addr){
	return;
      }
    }
    else{
      printf("XXX testFindPeer() node found on node list (but it has left) %s\n",
	     inet_ntoa(sin.sin_addr));
      assert(0);
    }
  }
  assert(0); /* Not found but should have been */
}


/*
 *------------------------------------------------------------------
 *
 * testPlax --
 *
 *          For a set of objects, calculate who are parent/children
 *          should be by hand and see if things match expected.
 *
 * Arguments:
 *      None.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------4.24.1998--------
 */
static void
testPlax(void)
{
  int iurl;
  char url[100];
  URLKey objKey;
  int otherMatch, maxMatch, myMatch;
  GenericKey otherKey;
  HCNLIter i;
  struct sockaddr_in sin, gotSin, *childrenA = NULL;
  int qNdx, amIRoot;
  int count = 0;
  int iAmLeaf, nChildren;

  /*
   * By now, we should have created a bunch of dummy objects and
   * therefore found out about neighbors' dummy objects and
   * more importantly found out about our neighbors existance.
   * So, HCHier_CheckIfChild should work.
   */
  for(HCNodeList_IterStart(&i);
      HCNodeList_IterCheck(i);
      HCNodeList_IterNext(&i)){
    if(HCNodeList_IterGetAddr(i, &sin)){
      /* Should be initialized from DNS by the time we get here */
      assert(sin.sin_addr.s_addr != 0);
      assert(HCHier_CheckIfChild(&sin));
      count++;
    }
  }

  if(count != 1){
    printf("Skipping HCHier::testPlax b/c test only works in 2-node system\n");
    return;
  }


  /*
   * Now, see if the system can figure out which nodes are our
   * parents and which are our children for different objects.
   */
  for(iurl = 0; iurl < 100; iurl++){
    sprintf(url, "http://www.cs.utexas.edu/%d", iurl);
    URLKey_Init(&objKey, url);
    
    myMatch = HCPlax_matchBits(HCPlax_myNodeKey(), objKey.key);
    maxMatch = myMatch;

    for(HCNodeList_IterStart(&i);
	HCNodeList_IterCheck(i);
	HCNodeList_IterNext(&i)){
      if(HCNodeList_IterGetAddr(i, &sin)){
	/* Should be initialized from DNS by the time we get here */
	assert(sin.sin_addr.s_addr != 0);
	otherKey = NodeKK_Key(&sin);
	otherMatch = HCPlax_matchBits(otherKey, objKey.key);
	if(otherMatch > maxMatch){
	  maxMatch = otherMatch;
	}

	if(otherMatch > myMatch){
	  if(otherMatch == maxMatch){
	    /*
	     * Parents 
	     */
	    assert(HCHier_CheckIfParent(objKey, &sin));
	    qNdx = HCHier_ParentQIndex(&objKey, &amIRoot);
	    assert(!amIRoot);
	    assert(!HCHier_GetParentAddr(qNdx, &gotSin));
	    assert(!SINCMP(&gotSin, &sin));
	    /*
	     * Children
	     */
	    qNdx = HCHier_ChildQIndex(&objKey, amIRoot, &iAmLeaf);
	    assert(iAmLeaf);
	  }
	  else{
	    assert(0); /* Tests work only for 2-node case right now */
	  }
	}
	else if(otherMatch == myMatch){
	  if(otherMatch == maxMatch){
	    if(HCPlax_myNodeKey() > otherKey){
	      /* 
	       * I am root
	       */
	      qNdx = HCHier_ParentQIndex(&objKey, &amIRoot);
	      assert(amIRoot);
	      qNdx = HCHier_ChildQIndex(&objKey, amIRoot, &iAmLeaf);
	      assert(!iAmLeaf);
	      nChildren = HCHier_GetChildAddrs(qNdx, &childrenA);
	      assert(nChildren == 1); /* Tests work only for 2-node case */
	      assert(!SINCMP(&sin, &childrenA[0]));
	      xfree(childrenA);
	    }
	    else{
	      /*
	       * Other node is root 
	       */
	      assert(HCHier_CheckIfParent(objKey, &sin));
	      qNdx = HCHier_ParentQIndex(&objKey, &amIRoot);
	      assert(!amIRoot);
	      assert(!HCHier_GetParentAddr(qNdx, &gotSin));
	      assert(!SINCMP(&gotSin, &sin));
	      qNdx = HCHier_ChildQIndex(&objKey, amIRoot, &iAmLeaf);
	      assert(iAmLeaf);
	    }
	  }
	  else{
	    assert(0); /* Tests work only for 2-node case right now */
	  }
	}
	else{
	  assert(otherMatch < myMatch);
	  if(myMatch == maxMatch){
	    /* 
	     * I am root (test only works in 2-node case)
	     */
	    qNdx = HCHier_ParentQIndex(&objKey, &amIRoot);
	    assert(amIRoot);
	    qNdx = HCHier_ChildQIndex(&objKey, amIRoot, &iAmLeaf);
	    assert(!iAmLeaf);
	    nChildren = HCHier_GetChildAddrs(qNdx, &childrenA);
	    assert(nChildren == 1); /* Tests work only for 2-node case */
	    assert(!SINCMP(&sin, &childrenA[0]));
	    xfree(childrenA);
	  }
	  else{
	    assert(0); /* Tests work only for 2-node case right now */
	  }
	}
      }
    }
  }
}


/*
 *------------------------------------------------------------------
 *
 * testCreateStuff --
 *
 *          Create and advertise a bunch of random objects. This is just
 *          to cause communication so that nodes learn about
 *          one another. We will never actually use these objects.
 *         
 *
 * Arguments:
 *      None.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------4.28.1998------
 */
static const int STUFF_COUNT = 1000;

static void 
testCreateStuff(char *tag)
{
  int ii;
  char name[128];

  for(ii = 0; ii < STUFF_COUNT; ii++){
    sprintf(name, "http://foo.www.cs.utexas.edu/self-test-hchier/%d/%s\n",
	    ii, tag); 
    HCinformLocalCopy(name);
  }
}


/*
 *------------------------------------------------------------------
 *
 * testFind --
 *
 *          See if a bunch of items are where we expect them to be.
 *
 * Arguments:
 *      None.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *---------------------------------------------------5.4.1998---------
 */
static void
testFind(void)
{
  char tag[256];
  struct sockaddr_in sin;
  HCNLIter i;
  
  /*
   * The local stuff should NOT be found in the local index
   * by convention.
   */
  sprintf(tag, "%s", inet_ntoa(getMyAddr()->sin_addr));
  testFindStuff(tag, NULL);

  /* 
   * Check to find the remote stuff
   */
  for(HCNodeList_IterStart(&i);
      HCNodeList_IterCheck(i);
      HCNodeList_IterNext(&i)){
    if(HCNodeList_IterGetAddr(i, &sin)){
      /* Should be initialized from DNS by the time we get here */
      assert(sin.sin_addr.s_addr != 0);
      sprintf(tag, "%s", inet_ntoa(sin.sin_addr));
      testFindStuff(tag, &sin);
    }
  }

  /*
   * Shared stuff should be local to everyone and therefore
   * not found
   */
  testFindStuff("shared", NULL);
}

/*
 *------------------------------------------------------------------
 *
 * testFindStuff --
 *
 *          See if a bunch of items are where we expect them to be.
 *
 * Arguments:
 *      None.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------5.4.1998-------
 */
static void
testFindStuff(char *tag, const struct sockaddr_in *expect)
{
  int ii;
  char name[128];
  struct sockaddr_in found;
  static float threshold = 0.05; /* Tolerate at most 5% collisions, etc */
  static int missCount = 0;

  for(ii = 0; ii < STUFF_COUNT; ii++){
    sprintf(name, "http://foo.www.cs.utexas.edu/self-test-hchier/%d/%s\n",
	    ii, tag); 
    if(HCfindNearest(name, &found) == NULL){
      if(expect != NULL){
	/*
	 * Miss could be from bug or could be due to limited 
	 * associativity and/or size. See how many misses we get
	 * and decide if that is too many
	 */
	printf("HCHier warning: unexpected miss (could be cache effect)\n");
	missCount++;
      }
    }
    else{
      if(expect == NULL){
	assert(0); /* found something we didn't expect to find */
      }
      else if(SINCMP(expect, &found)){
	assert(expect != NULL);
	/* 
	 * Miss - found data but not where expected.
	 */
	assert(0);
      }
    }
  }
  if(missCount > threshold * STUFF_COUNT){
    /* too many misses */
    assert(0);
  }
  missCount = 0;
}
