/*
 *------------------------------------------------------------------
 *
 * LESS Group 
 * Computer Sciences Department 
 * University of Texas at Austin
 *
 * HCHier ---
 *          Abstraction of the logical hierarchy of nodes.
 *Basic idea of the interface: the network code works by enqueuing sets
 *of messages for its parent(s) and children. The hierarchy manages the
 *mapping from an array of queues to the parents and children in the
 *hierarchy. The hierarchy tells the network how many queues to
 *maintain, and for each message the network code asks the hierarchy for
 *the index of the queue to enqueue the message. Then, when it is
 *finally time to send the messages in the queues, the network asks the
 *hierarchy for the address to which to send the data in a queue.
 *
 *The rest of the system only knows of the queues by their index; at
 *enqueue time the hierarhcy encapsulates the mapping from logical
 *destination (e.g. "my parent for objects with id 0xFF3028FA"-->
 *"enqueue in buffer 32"; "my level-2 children" --> "enqueue in buffer
 *17") to a buffer index, and at send time the hierarchy encapsulates
 *the mapping from index to physical destination (e.g. "who is index 32"
 *--> "128.32.111.41:8081")
 *
 *One important point -- you should never cache the mapping from
 *index-->physical address outside of HCHier. The reason is that we want
 *this mapping to change when new nodes are discovered or old nodes die.
 *HCHier is responsible for maintaining the dynamic mapping.
 *
 * $Date: 1998/05/04 18:42:21 $ $Id: HCHier.h,v 1.8 1998/05/04 18:42:21 dahlin Exp $
 *
 * Copyright (c) 1997 by Mike Dahlin.
 * 
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for noncommercial use without fee and
 * without written agreement is hereby granted, provided that 
 * the above copyright notice appear in all copies of this software.
 * 
 *------------------------------------------------------------------
 */
#ifndef _HCHier_h_
#define _HCHier_h_

extern int HCHier_selfTestDone;

void HCHier_Init _PARAMS((void));
void HCHier_Destroy _PARAMS((void));

/*
 * Interface to HCHint cache policy in disk -- need to 
 * know who is closer/farther to decide which hints to keep.
 *
 * Return < 0 if node1 is closer
 *        0 if the same
 *        > 0 if node2 is closer
 */
int HCHier_compareDistanceFromMe _PARAMS((struct in_addr new, 
                                        struct in_addr old));

/*
 * Interface to join/leave so we can keep track of
 * what nodes exist in the system
 */
void HCHier_newNode _PARAMS((struct sockaddr_in *sin, int hops, 
			     long long time));
void HCHier_delNode _PARAMS((struct sockaddr_in *sin, int hops,
			     long long time));


/*
 * We also need to help figure out the hierarchy for Plaxton's
 * algorithm. Plaxton's algorithm figures out who the parents
 * are, but it doesn't know who the children are. We observe
 * messages to keep in sync. 
 * On recv message:
 *	recv parent msg from parent: process parent message
 *	recv child msg from child: process child message
 *	recv parent msg from !parent: send "I am not child" to parent;
 *			drop message
 *	recv child msg from !child: add child to child list;
 *				process child message
 *	recv "I am not child" from child: remove child
 */
int HCHier_CheckIfParent _PARAMS((URLKey key, struct sockaddr_in *candidate));
int HCHier_CheckIfChild _PARAMS((struct sockaddr_in *candidate));
void HCHier_AddChild _PARAMS((struct sockaddr_in *child));
void HCHier_RemoveChild _PARAMS((struct sockaddr_in *child));


/*
 * Network enqueue/send interface.
 */

/*  Each children queue holds messages that should go to a set of
 * "equivalent" children  (eg. the children at some level of the
 * hierarchy.)  There are HCHier_NChildrenQs() such queues;
 * Child queue q contains c = HCHier_ChildrenPerQ(q) children.
 * The addresses of the children can be found with the function
 * HCHier_GetChildAddr(q, c);
 */
int HCHier_NChildrenQs _PARAMS((void));
int HCHier_ChildQIndex _PARAMS((URLKey *key, int iAmRoot, int *amILeaf));
int HCHier_GetChildAddrs _PARAMS((int qIndex, struct sockaddr_in **retAddr));

/*
 * Each parent queue holds messages that should go to one parent.
 * In general, we may have different parents for different objects.
 * So, enqueue messages for an object in the queue
 * HCHier_ParentQIndex(URLKey).
 */
int HCHier_NParentQs _PARAMS((void));
int HCHier_ParentQIndex _PARAMS((URLKey *key, int *amIRoot));
int HCHier_GetParentAddr _PARAMS((int index, struct sockaddr_in *retAddr));

void HCHier_selfTest _PARAMS((void));
void HCHier_testNetwork _PARAMS((void));

#ifdef DOTEST
void HCHier_selfTest _PARAMS((void));
#endif

#endif _HCHier_h_

