/*
 *------------------------------------------------------------------
 *
 * LESS Group 
 * Computer Sciences Department 
 * University of Texas at Austin
 *
 * HCNodelist ---
 * The nodelist keeps a list of all nodes in the hierarchy whether they
 * are alive or dead. It uses a flooding algorithm for this (see 
 * ../notes.3.17.1998). As part of this flooding algorithm, it learns
 * a hopcount to other nodes. We tell squid_db this hopcount as an
 * initial *rough* estimate of distance between nodes.
 *  
 * The flooding algorithm needs as input a list of "neighbors". The lists
 * of neighbors for all nodes MUST overlap. The lists of neighbors SHOULD
 * include only nearby neighbors (so that the hopcounts are correct). 
 * It uses the squid static hierarchy (see neighbors.c e.g. getFirstPeer()
 * etc) for this initial list.
 *
 * $Date: 1998/05/04 18:42:46 $ $Id: HCNodelist.c,v 1.5 1998/05/04 18:42:46 dahlin Exp $
 *
 * Copyright (c) 1997 by Mike Dahlin.
 * 
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for noncommercial use without fee and
 * without written agreement is hereby granted, provided that 
 * the above copyright notice appear in all copies of this software.
 * 
 *------------------------------------------------------------------
 */
#include <assert.h>

#define FRIEND_OF_HCNODELIST 1 /* So we can see .h file */
#define FRIEND_OF_HCPLAX 1
#include "squid.h"

typedef struct HierNodeElem{
  List_Links links;  /* MUST be first element of structure */
  struct sockaddr_in sin;
  int action; /* HC_Join or HC_Leave */
  long long timestamp_ns; /* time of action */
}HierNodeElem;

static List_Links nodes;
static int myStatus = HC_Leave;

static void HierNodeElem_Init _PARAMS((HierNodeElem *, struct sockaddr_in *,
				       int action, long long timestamp_ns));

static HierNodeElem *nodeFind(struct sockaddr_in *sin);
static long long currentTimeNS();
static void floodEvent(int action, const struct sockaddr_in *source, 
		       long long timeNS, int hopcount);

static void HCNodelist_OutcallHops(struct sockaddr_in *sin, 
				    int hops, long long timeNS);
static void HCNodelist_OutcallJoin(struct sockaddr_in *sin, 
				    int hops, long long timeNS);
static void HCNodelist_OutcallLeave(struct sockaddr_in *sin, 
				    int hops, long long timeNS);






static void
HierNodeElem_Init(HierNodeElem *e, struct sockaddr_in *sin, int action,
		  long long timestamp_ns)
{
  assert(action == HC_Join || action == HC_Leave);
  List_InitElement(&e->links);
  e->sin = *sin;
  e->action = action;
  e->timestamp_ns = timestamp_ns;
}

/*
 *------------------------------------------------------------------
 *
 * HCNodeList_Init --
 *
 *          Constructor.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      Initialize nodelist to empty list.
 *
 *----------------------------------------------3.31.1998-------------
 */
void 
HCNodelist_Init (void)
{
  List_Init(&nodes);
}

/*
 *------------------------------------------------------------------
 *
 * HCNodeList_Destroy --
 *
 *          Deallocate all elements in list
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------3.31.1998-------
 */
void 
HCNodelist_Destroy (void)
{
  List_Links *item;
  while(!List_IsEmpty(&nodes)){
    item = List_First(&nodes);
    List_Remove(item);
    free(item);
  }
}


/*
 *------------------------------------------------------------------
 *
 * HCNodeList_LocalJoin --
 *
 *          Do a local join. This tells the algorithm
 *          to flood "join" messages to the neighbors.
 *
 * Arguments:
 *      struct sockaddr_in myId -- my ip address
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------4.1.1998----
 */
void 
HCNodelist_LocalJoin(void)
{
  const struct sockaddr_in *myId;
  int hopcount;
  long long timeNS;

  myStatus = HC_Join;

  myId = getMyAddr();
  timeNS = currentTimeNS();
  hopcount = 1;
  floodEvent(HC_Join, myId, timeNS, hopcount);
}

/*
 *------------------------------------------------------------------
 *
 * HCNodeList_LocalLeave --
 *
 *          Do a local leave. This tells the algorithm
 *          to flood "leave" messages to the neighbors.
 *
 * Arguments:
 *      struct sockaddr_in myId -- my ip address
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------4.1.1998----
 */
void 
HCNodelist_LocalLeave(void)
{
  const struct sockaddr_in *myId;
  int hopcount;
  long long timeNS;

  myStatus = HC_Leave;
  myId = getMyAddr();
  timeNS = currentTimeNS();
  hopcount = 1;
  floodEvent(HC_Leave, myId, timeNS, hopcount);
}


/*
 *------------------------------------------------------------------
 *
 * HCNodeList_getMyStatus --
 *
 *          Am I currently in or out?
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------4.1.1998--------
 */
int
HCNodelist_myStatus(void)
{
  return myStatus;
}


/*
 *------------------------------------------------------------------
 *
 * HCNodeList_NetJoin --
 *
 *          Process a "join" received from the network.
 *          If this is a new node, a new action for a node,
 *          or a current action with a shorter distance,
 *          proceed with the join and propagate message
 *          to neighbors.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      Allocates and adds an element to the list; calls 
 *      OutcallJoin().
 *
 *------------------------------------------------------3.31.1998----
 */
void 
HCNodelist_NetJoin(struct sockaddr_in *sin, int hops, long long joinTimeNS)
{
  HierNodeElem *h, *oldCopy;
  int oldHops;
  char name[16];
  int foundOldJoin = 0;

  assert(sin);
  /* Ignore if it's already on list and either (1) a newer record
   * or (2) a current record and at least as close as this
   * update.  Also ignore if it is me!
   * For simplicity, rather than update a record in place, remove
   * it and add a new record to front of list. This should be better
   * for performance, too, since we expect to get a burst of
   * notifications from all neighbors, so put record at front of list. 
   */
  if(!SINCMP(sin, getMyAddr())){
    return;
  }
  oldCopy = nodeFind(sin);
  if (oldCopy != NULL){
    if(oldCopy->timestamp_ns > joinTimeNS){
      /* We got a stale message; ignore it */
      return;
    }
    assert(oldCopy->timestamp_ns <= joinTimeNS);
    if(oldCopy->action == HC_Join){
      foundOldJoin = 1;
    }
    if(oldCopy->timestamp_ns < joinTimeNS){
      List_Remove((List_Links *)oldCopy);
      free(oldCopy);
    }
    else{
      /* Action must be join: else had join and Leave at same time. */
      assert(oldCopy->timestamp_ns == joinTimeNS);
      assert(oldCopy->action == HC_Join); 
      strcpy(name, inet_ntoa(sin->sin_addr));
      oldHops = netdbHostHops(name);
      if(oldHops > hops){
	/* New join supercededs */
	List_Remove((List_Links *)oldCopy);
	free(oldCopy);
      }
      else{
	/* Old join survives */
	return;
      }
    }
  }
  /*
   * Only get here if new join supercedes and old join removed or
   * nonexistant 
   */
  assert(!nodeFind(sin));

  h = (HierNodeElem *)xmalloc(sizeof(HierNodeElem));
  HierNodeElem_Init(h, sin, HC_Join, joinTimeNS);
  List_Insert((List_Links *)h, &nodes);

  /*
   * Now tell everyone else (locally and network) about 
   * the new addition or new distance
   */
  if(foundOldJoin){
    HCNodelist_OutcallHops(sin, hops, joinTimeNS);
  }
  else{
    HCNodelist_OutcallJoin(sin, hops, joinTimeNS);
  }
}

/*
 *------------------------------------------------------------------
 *
 * HCNodelist_NetLeave --
 *
 *          Process a notification that a node has left the system
 *
 *          We keep "departed" nodes on nodelist with a "leave" flag
 *          set so we know when to stop propagating messages.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------3.31.1998--------------
 */
void 
HCNodelist_NetLeave (struct sockaddr_in *sin, int hops, 
		     long long leaveTimeNS)
{
  HierNodeElem *h, *oldCopy;
  int foundOldLeave = 0;

  assert(sin);

  /* Ignore if it's already on list and a newer record.
   * For simplicity, rather than update a record in place, remove
   * it and add a new record to front of list. This should be better
   * for performance, too, since we expect to get a burst of
   * notifications from all neighbors, so put record at front of list. 
   */
  oldCopy = nodeFind(sin);
  if (oldCopy){
    if(oldCopy->action == HC_Leave){
      foundOldLeave = 1;
    }
    if(oldCopy->timestamp_ns < leaveTimeNS){
      List_Remove((List_Links *)oldCopy);
      free(oldCopy);
    }
    else{
      /* 
       * old copy is actually newer than this update. Ignore this
       * update.
       */
      return;
    }
  }
  /*
   * Only get here if new leave supercedes old record removed or
   * nonexistant 
   */
  assert(!nodeFind(sin));

  h = (HierNodeElem *)xmalloc(sizeof(HierNodeElem));
  HierNodeElem_Init(h, sin, HC_Leave, leaveTimeNS);
  List_Insert((List_Links *)h, &nodes);

  /*
   * Now tell everyone else (locally and network) about 
   * the new addition or new distance
   */
  if(!foundOldLeave){
    HCNodelist_OutcallLeave(sin, hops, leaveTimeNS);
  }
  
}

/*
 *------------------------------------------------------------------
 *
 * nodeFind --
 *
 *          Find node with matching address.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
static HierNodeElem *
nodeFind(struct sockaddr_in *sin)
{
  HierNodeElem *e;
  List_Links *item;
  
  LIST_FORALL(&nodes, item) {
    e = (HierNodeElem *) item;
    if (!SINCMP(&e->sin, sin)){
      return(e);
    }
  }
  return(NULL);
}

/*
 *------------------------------------------------------------------
 *
 * floodEvent --
 *
 *          Send the notification to our neighbors.
 *
 * Arguments:
 *      int action HC_Join or HC_Leave
 *      struct sockaddr_in *source -- the subject of the action
 *      long long timeNS -- the time of the action (used for
 *                          ordering events and removing duplicates)
 *      int hopcount -- how many hops has this event seen?
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *-------------------------------------------------3.31.1998--------
 */
static void
floodEvent(int action, const struct sockaddr_in *source, long long timeNS, 
	   int hopcount)
{
  HCEntry entry;
  peer *p;
  URLKey hackURLKey;
  struct sockaddr_in dest;

  assert(action == HC_Join || action == HC_Leave);

  if(neighborsOutgoing == NULL){
    return;
  }

  hackURLKey.key = timeNS; /* XXX Update should be a union XXX */

  HCEntry_Init(&entry, hackURLKey, source);
  HCNet_Enqueue(neighborsOutgoing, action, &entry, hopcount);
  HCNet_Complete(neighborsOutgoing);

  for (p = getFirstPeer(); p; p = getNextPeer(p)) {
    if(p->in_addr.sin_addr.s_addr == 0){
      /* printf("XXX HCNodelist warning: peer %s addr not set\n", 
	 p->host);*/
      continue;
    }
    dest.sin_family = AF_INET;
    dest.sin_addr = p->in_addr.sin_addr;
    dest.sin_port = p->http_port;
    HCNet_SendTo(neighborsOutgoing, &dest);
  }

  HCNet_Done(neighborsOutgoing);
}





/*
 *------------------------------------------------------------------
 *
 * HCNodeList_IterXXX --
 *
 *          Used for iterating through all addresses on list.
 *
 *   for(HCNodeList_IterStart(&iter);
 *       HCNodeList_IterCheck(&iter);
 *       HCNodeList_IterNext(&iter)){
 *     valid = HCNOdeList_IterGetAddr(&iter, &addr);
 *   }
 *  
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      Note -- only itereates through "joined" elements;
 *      Skip the shadow elements representing nodes that
 *      have left.
 *
 *      Warning -- iterator is pointer into element on list.
 *      If list changing during iteration, could get confused.
 *
 *------------------------------------------------------------------
 */
void 
HCNodeList_IterStart(HCNLIter *iter)
{
  *iter = (HCNLIter)List_First(&nodes);
}

int
HCNodeList_IterCheck(HCNLIter iter)
{
  return(!List_IsAtEnd(&nodes, iter));
}

void 
HCNodeList_IterNext(HCNLIter *iter)
{
  HierNodeElem *elem;

  if(List_IsAtEnd(&nodes, *iter)){
    return;
  }

  while(1){
    *iter = (HCNLIter)List_Next(*iter);
    /*
     * Skip non-HC_Join elements by keeping going in that case.
     */
    if(List_IsAtEnd(&nodes, *iter)){
      return;
    }
    elem = (HierNodeElem *)*iter;
    if(elem->action == HC_Join){
      return;
    }
  }
    
}

/*
 *------------------------------------------------------------------
 *
 * HCNodeList_IterGetAddr --
 *
 *          description.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
int
HCNodeList_IterGetAddr(HCNLIter iter, struct sockaddr_in *ret)
{
  HierNodeElem *elem;
  elem = (HierNodeElem *)iter;
  *ret = elem->sin;
  if(elem->action == HC_Join){
    return 1;
  }
  return 0;
}



/*
 *------------------------------------------------------------------
 *
 * HCNodelist_OutcallHops --
 *
 *          This function gets called whenever the number of hops
 *          to a node *changes* due to network updates. This
 *          is a good place to put function calls to nodes that
 *          care to hear about such things.
 *
 *          Right now, 3 people care:
 *          1) tell neighbors via flooding
 *          2) netdb
 *          3) plaxton dynamic hierarchy algorithm
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
static void 
HCNodelist_OutcallHops(struct sockaddr_in *sin, int hops, long long timeNS)
{
  char name[16];

  /*
   * Tell neighbors
   */
  floodEvent(HC_Join, sin, timeNS, hops+1);

  /*
   * Tell netdb
   */
  strcpy(name, inet_ntoa(sin->sin_addr));
  netdbAdd(sin->sin_addr, name);
  netdbHandlePingReply(sin, hops, 0);

  /* 
   * Tell plaxton algorithm
   */
  HCPlax_changeDistance(sin);
} 

/*
 *------------------------------------------------------------------
 *
 * HCNodelist_OutcallJoin --
 *
 *          This function gets called whenever a node joins the
 *          sytsem.
 *          This is a good place to put function calls to nodes that
 *          care to hear about such things.
 *
 *          Right now, 3 people care:
 *          1) tell neighbors via flooding
 *          2) netdb
 *          3) plaxton dynamic hierarchy algorithm
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
static void 
HCNodelist_OutcallJoin(struct sockaddr_in *sin, int hops, long long timeNS)
{
  char name[16];

  /*
   * Tell neighbors
   */
  floodEvent(HC_Join, sin, timeNS, hops+1);

  /*
   * Tell netdb
   */
  strcpy(name, inet_ntoa(sin->sin_addr));
  netdbAdd(sin->sin_addr, name);
  netdbHandlePingReply(sin, hops, 0);

  /* 
   * Tell plaxton algorithm
   */
  HCPlax_addNode(sin);
}

/*
 *------------------------------------------------------------------
 *
 * HCNodelist_OutcallLeave --
 *
 *          This function gets called whenever a node leaves the
 *          sytsem.
 *          This is a good place to put function calls to nodes that
 *          care to hear about such things.
 *
 *          Right now, 2 people care:
 *          1) tell neighbors via flooding
 *          2) plaxton dynamic hierarchy algorithm
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
static void 
HCNodelist_OutcallLeave(struct sockaddr_in *sin, int hops, long long timeNS)
{
  /*
   * Tell neighbors
   */
  floodEvent(HC_Join, sin, timeNS, hops);

  /* 
   * Tell plaxton algorithm
   */
  HCPlax_removeNode(sin);
}


/*
 *------------------------------------------------------------------
 *
 * currentTimeNS --
 *
 *          Return the current time in nanoseconds since Jan 1, 1970.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
static long long
currentTimeNS()
{
  static const long long NS_PER_SEC = 1000000000;
  static const long long NS_PER_USEC = 1000;
  int error;
  struct timeval time;
  long long ret;

  error = gettimeofday(&time, NULL);
  assert(error != -1);
  ret = NS_PER_SEC * (long long)time.tv_sec 
    + NS_PER_USEC * (long long)time.tv_usec;
  return ret;
}


/*
 *------------------------------------------------------------------
 *
 * HCNodelist_SelfTest --
 *
 *          description.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
void HCNodelist_SelfTest (void)
{
  struct sockaddr_in n0, n1, n2;
  HCNLIter iter;
  struct sockaddr_in nodeAddr;
  int nhops;
  char name[16];
  int valid;
  int found = 0;

  printf("HCNodelist_SelfTest...");
  n0.sin_addr.s_addr = 0x90536F00; n0.sin_port = 0;
  n1.sin_addr.s_addr = 0x91536F01; n1.sin_port = 1;
  n2.sin_addr.s_addr = 0x92536F02; n2.sin_port = 2;

  HCNodelist_Init();
  HCPlax_Init(8);

  assert(neighborsOutgoing == NULL);
  neighborsOutgoing = (HCNet *)malloc(sizeof(HCNet));
  assert(neighborsOutgoing);
  HCNet_Init(neighborsOutgoing);
  HCNet_SetTestMode(neighborsOutgoing);


  /*
   * How to test if this really works? For now look at how many
   * outgoing messages were enqueued.
   */
  HCNodelist_LocalJoin(); 
  HCNodelist_LocalLeave();
  assert(HCNet_GetTestCount(neighborsOutgoing) == 2);

  /*
   * Expected final state:
   * n0 JOIN dist = 12
   * n1 LEAVE XXX
   * n2 JOIN dist=20
   */
  HCNodelist_NetJoin(&n0, 10, 42);
  HCNodelist_NetJoin(&n2, 20, 42);
  HCNodelist_NetJoin(&n0, 13, 43);
  HCNodelist_NetJoin(&n1, 20, 42);
  HCNodelist_NetJoin(&n2, 20, 100);
  HCNodelist_NetJoin(&n0, 9, 42);
  HCNodelist_NetJoin(&n0, 12, 43);
  HCNodelist_NetJoin(&n2, 20, 42);
  HCNodelist_NetLeave(&n1, 999, 43);
  HCNodelist_NetJoin(&n1, 19, 42);

  for(HCNodeList_IterStart(&iter);
      HCNodeList_IterCheck(iter);
      HCNodeList_IterNext(&iter)){
    valid = HCNodeList_IterGetAddr(iter, &nodeAddr);
    if(valid){
      strcpy(name, inet_ntoa(nodeAddr.sin_addr));
      nhops = netdbHostHops(name);
      if(nodeAddr.sin_addr.s_addr == n0.sin_addr.s_addr){
	found++;
	assert(nhops == 12);
      }
      if(nodeAddr.sin_addr.s_addr == n2.sin_addr.s_addr){
	found++;
	assert(nhops = 20);
      }
      if(nodeAddr.sin_addr.s_addr == n1.sin_addr.s_addr){
	assert(0);
      }
    }
  }
  assert(found == 2);

  HCNodelist_Destroy();
  HCPlax_Destroy();

  HCNet_Destroy(neighborsOutgoing);
  free(neighborsOutgoing);
  neighborsOutgoing = NULL;

  printf("..OK\n");

}

