/*
 *------------------------------------------------------------------
 *
 * LESS Group 
 * Computer Sciences Department 
 * University of Texas at Austin
 *
 * HCPrimitives.h ---
 *          Primitive types for hint cache.
 *
 * $Date: 1998/05/06 21:15:33 $ $Id: HCPrimitives.c,v 1.14 1998/05/06 21:15:33 dahlin Exp $
 *
 * Copyright (c) 1997 by Mike Dahlin.
 * 
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for noncommercial use without fee and
 * without written agreement is hereby granted, provided that 
 * the above copyright notice appear in all copies of this software.
 * 
 *------------------------------------------------------------------
 */
#include "squid.h"
#include <assert.h>


const URLKey INVALID_URL_KEY = {0};

static GenericKey calculateStupidKey(char *url);
static GenericKey calculateMD5Key(char *url);
static void makeLowerCase(char *url);
static GenericKey calculateKey(char *url);

static int updateSequenceNumber = 0;
const static int useMD5 = 1;


/*
 *------------------------------------------------------------------
 *
 * URLKey --
 *
 *          Constructor. Calculate a hash of the URL.
 *          This hash needs to be a good one; collisions
 *          must be rare because they get treated as misses
 *          or worse yet false hits.
 *
 *          Be sure to convert the URL to a canonical form.
 *          We assume that the only canonicalizing that
 *          we need to do is making it all the same case.
 *
 * Arguments:
 *      char *url -- a URL.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */

void
URLKey_Init(URLKey *key, char *url)
{
  assert(key);
  if(!url){
    key->key = 0;
    return;
  }

  makeLowerCase(url);
  key->key = calculateKey(url);    
}


int
URLKey_Compare(const URLKey *k1, const URLKey *k2)
{
  assert(k1);
  assert(k2);
  if(k1->key != k2->key){
    return 1;
  }
  return 0;
}


/*
 * Make lower case, but stop at third '/' (e.g., after the hostname).
 */
static void
makeLowerCase(char *url)
{
  int len, ii;
  int slashCount = 0;

  assert(url);
  len = strlen(url);
  for(ii = 0; ii < len; ii++){
    /*
     * Alignment.
     */
    if(url[ii] == '/'){
      slashCount++;
      if(slashCount == 3){
	return;
      }
    }
    url[ii] = tolower(url[ii]);
  }
}


/*
 * The hash function should be really good since we are counting
 * on there being few collisions in our k-way associative
 * table (for small k). The roll-your-own one I doubt is much
 * good, so you should probably use the MD5 hash. Notice
 * that machines send messages to each other that contain
 * the hash only, so all machines must use the same calculateKey
 * function.
 */
static GenericKey
calculateKey(char *url)
{

  if(useMD5){
    return calculateMD5Key(url);
  }
  else{
    return calculateStupidKey(url);
  }

}


static GenericKey
calculateMD5Key(char *url)
{
  MD5_CTX  tCrypt;
  static long long unsigned digestPart;

  assert(sizeof(long long unsigned) == sizeof(GenericKey));
  assert(url);
  MD5Init(&tCrypt);
  MD5Update(&tCrypt, "this is my secret key, yeah right", strlen("this is my secret key, yeah right"));
  MD5Update(&tCrypt, url, strlen(url));
  MD5Final(&tCrypt);
  /*
   * The digest MUST be smaller than a long long (otherwise we
   * would be including undefined bytes past end of digest.
   */
  assert(MD5_DIGEST_SIZE * sizeof(unsigned char) 
	 >= sizeof(long long unsigned));
				   
  memcpy(&digestPart, &(tCrypt.digest[0]), sizeof(digestPart));
  return digestPart;
}

static GenericKey
calculateStupidKey(char *url)
{
  static const long long unsigned BIG_PRIME_1 = 193844587;
  static const long long unsigned BIG_PRIME_2 = 455899;
  static const long long unsigned BIG_PRIME_3 = 901367;
  
  int len, ii;
  long long unsigned key, chunk;
  assert(sizeof(long long unsigned) == sizeof(GenericKey));
  assert(url);
  len = strlen(url);
  key = 0;
  for(ii = 0; ii < len-8; ii += 8){
    /* Make copy to guarantee alignment. */
    memcpy(&chunk, &url[ii], 8);
    key = key + chunk * BIG_PRIME_1;
  }
  for( ; ii < len; ii++){
    key = key + (unsigned char)url[ii] * BIG_PRIME_2;
  }
  key = key + len * BIG_PRIME_3;
  return key;
}


void 
Net_URLKey_Init(Net_URLKey *mungedKey, URLKey *key)
{
  mungedKey->mungedKey = HCEndian_llMachineToNet(key->key);
}

void
Net_URLKey_LocalFormat(Net_URLKey *mungedKey, URLKey *key)
{
  key->key = HCEndian_llNetToMachine(mungedKey->mungedKey);
}



/*
 *------------------------------------------------------------------
 *
 * NodeKey --
 *
 *          A <key, struct sockaddr_in> pair (plus stuff to
 *          put it on a list.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *---------------------------------------------------4.9.1998--------
 */
void
NodeKey_Init(NodeKey *k, const struct sockaddr_in *addrP)
{
  assert(k);
  assert(addrP);
  
  List_Init((List_Links *)k);
  k->key = NodeKK_Key(addrP);
  k->addr = *addrP;
}

int
NodeKey_Compare(const NodeKey *k1, const NodeKey *k2)
{
  assert(k1);
  assert(k2);
  if(k1->key != k2->key){
    return 1;
  }
  return 0;
}

struct sockaddr_in 
NodeKey_GetAddr (const NodeKey *k)
{
  assert(k);
  return k->addr;
}

GenericKey
NodeKK_Key(const struct sockaddr_in *addrP)
{
  GenericKey k;
  char *name;

  name = inet_ntoa(addrP->sin_addr);
  if(useMD5){
    k = calculateMD5Key(name);
  }
  else{
    k = calculateStupidKey(name);
  }
  return k;
}



/*
 *------------------------------------------------------------------
 *
 * HCEntry --
 *
 *          A structure including a <URLKey, sockaddr_in> pair. Generally
 *          points to the nearest copy of the specified url.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
void 
HCEntry_Init(HCEntry *entry, URLKey key, const struct sockaddr_in *saddr)
{
  entry->key = key;
  entry->ipaddr = saddr->sin_addr;
  entry->port = saddr->sin_port;
}

int
HCE_VALID(HCEntry *hce)
{
  if((hce->ipaddr.s_addr != INADDR_ANY)
     && URLKey_Compare(&hce->key, &INVALID_URL_KEY)){
    return 1;
  }
  return 0;
}

void
HCE_INVALIDATE(HCEntry *hce) 
{
  hce->ipaddr.s_addr  = INADDR_ANY;
  hce->key = INVALID_URL_KEY;
  return;
}

int
HCEntry_Compare(HCEntry *e1, HCEntry *e2)
{
  /* 
   * If you add a field, you must change this code.
   */
  if(URLKey_Compare(&e1->key, &e2->key)){
    return 1;
  }

  if (e1->ipaddr.s_addr != e2->ipaddr.s_addr)
      return 1;

  if (e1->port != e2->port)
      return 1;

  return 0;
}

void
Net_HCEntry_Init(Net_HCEntry *mungedEntry, HCEntry *entry)
{
  /* 
   * If you add a field, you must change this code.
   */
  Net_URLKey_Init(&mungedEntry->mungedKey, &entry->key);
  mungedEntry->mungedIpaddr.s_addr  = htonl(entry->ipaddr.s_addr);
  mungedEntry->mungedPort = htons(entry->port);
}

void
Net_HCEntry_LocalFormat(Net_HCEntry *mungedEntry, HCEntry *entry)
{
  /* 
   * If you add a field, you must change this code.
   */
  Net_URLKey_LocalFormat(&mungedEntry->mungedKey, &entry->key);
  entry->ipaddr.s_addr = ntohl(mungedEntry->mungedIpaddr.s_addr);
  entry->port = ntohs(mungedEntry->mungedPort);
}


/*
 *------------------------------------------------------------------
 *
 * HCUpdate --
 *
 *          An update is a HCEntry and an action.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
void
HCUpdate_Init(HCUpdate *u, int action, HCEntry *entry, char hopcount)
{
  assert(hopcount >= 0);

  u->action = action;
  u->len = sizeof(*u);
  u->hopcount = hopcount;
  u->seqno = updateSequenceNumber++;
  u->cchange = 0;
  u->entry = *entry;
}

void
HCUpdate_selfTest()
{
  /* Make sure net version of struct stays in sync. */
  /*  assert(sizeof(HCUpdate) == 28); */

  assert(sizeof(HCUpdate) == sizeof(Net_HCUpdate));
}

void
Net_HCUpdate_Init(Net_HCUpdate *mupdate, HCUpdate *update)
{
    /* No need for htonl on action, hopcount, or len - they are single bytes. */
    mupdate->mungedAction   = update->action;
    mupdate->mungedLen      = update->len;
    mupdate->mungedSeqno    = update->seqno;
    mupdate->mungedHopcount = update->hopcount;
    mupdate->mungedCchange  = htonl(update->cchange);
    
    Net_HCEntry_Init(&mupdate->mungedEntry, &update->entry);
}


void
Net_HCUpdate_LocalFormat(Net_HCUpdate *mupdate, HCUpdate *update)
{
    /* No need for htonl on action or len - they are single bytes. */
    update->action   = mupdate->mungedAction;
    update->len      = mupdate->mungedLen;
    update->seqno    = mupdate->mungedSeqno;
    update->hopcount = mupdate->mungedHopcount;
    update->cchange  = ntohl(mupdate->mungedCchange);
    
    Net_HCEntry_LocalFormat(&mupdate->mungedEntry, &update->entry);
}

int 
HCUpdate_Compare (HCUpdate *u1, HCUpdate *u2)
{
  /* must change this if add new fields */
  /*  assert(sizeof(HCUpdate) == 28); */

  if((u1->action != u2->action)
     || (u1->len != u2->len)
     || (u1->hopcount != u2->hopcount)
     || (u1->seqno != u2->seqno)
     || (HCEntry_Compare(&u1->entry, &u2->entry))){
    return 1;
  }
  return 0;
}

/*
 * Assume that ptr is a pointer to an unaligned bunch of data that
 * may be interpreted as an HCNetUpdate struct. Pull out the 
 * length field (an unsigned char), convert it to local machine
 * format (no op for an unsigned char), and return it as
 * an int.
 */
int 
HCNetUpdate_GrabLen(char *ptr)
{
  int len;
  len = ((Net_HCUpdate *)ptr)->mungedLen;
  return len;
}

#ifdef DOTEST
/*
 *------------------------------------------------------------------
 *
 * primitivesSelfTest --
 *
 *          description.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
void
primitivesSelfTest()
{

  URLKey urlKey, urlKey2;
  Net_URLKey netUrlKey;
  char url[128];
  HCEntry hcEntry, hcEntry2;
  Net_HCEntry netHCEntry;
  HCUpdate hcUpdate, hcUpdate2;
  Net_HCUpdate netHCUpdate;
  struct sockaddr_in saddr = { AF_INET, 0, { { { 0, 0, 0, 0 } } } };

  printf("primitivesSelfTest()..");

  sprintf(url, "http://www.cs.Utexas.edu");
  URLKey_Init(&urlKey, url);
  Net_URLKey_Init(&netUrlKey, &urlKey);
  Net_URLKey_LocalFormat(&netUrlKey, &urlKey2);
  assert(!URLKey_Compare(&urlKey, &urlKey2));

  HCEntry_Init(&hcEntry, urlKey, &saddr);
  Net_HCEntry_Init(&netHCEntry, &hcEntry);
  Net_HCEntry_LocalFormat(&netHCEntry, &hcEntry2);
  assert(!HCEntry_Compare(&hcEntry, &hcEntry2));

  HCUpdate_Init(&hcUpdate, HC_InvalToChild, &hcEntry, 1);
  HCUpdate_selfTest();
  Net_HCUpdate_Init(&netHCUpdate, &hcUpdate);
  Net_HCUpdate_LocalFormat(&netHCUpdate, &hcUpdate2);
  assert(!HCUpdate_Compare(&hcUpdate, &hcUpdate2));

  printf(".OK\n");
}
#endif /* DOTEST */
