/*
 *------------------------------------------------------------------
 *
 * LESS Group 
 * Computer Sciences Department 
 * University of Texas at Austin
 *
 * HCProp ---
 *          Algorithms for propagating hints.
 *
 * $Date: 1998/05/04 20:43:00 $ $Id: HCProp.c,v 1.4 1998/05/04 20:43:00 dahlin Exp $
 *
 * Copyright (c) 1997 by Mike Dahlin.
 * 
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for noncommercial use without fee and
 * without written agreement is hereby granted, provided that 
 * the above copyright notice appear in all copies of this software.
 * 
 *------------------------------------------------------------------
 */
#include <assert.h>
#include "squid.h"

/*
 *------------------------------------------------------------------
 *
 * hcprop_localAction --
 *
 *          What to do on a local invalidate or inform.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
void
hcprop_localAction(char *url, int action)
{
  HCEntry myentry;
  URLKey key;
  HCUpdate update;

  if (!HC_ACTIVE())
      /* Hint cache is inoperative.  Nothing to do. */
      return;
  
  assert(action == HC_InformToParent || action == HC_InvalToParent);
  
  if (!strncmp(url, "route", 5))
      /* Routes handle their own updates */
      /* XXX - can we get at the protocol_t for faster check? */
      return;
  
  URLKey_Init(&key, url);

  /*
   * Update local and remote hint caches.
   */
  HCEntry_Init(&myentry, key, (struct sockaddr_in *) getMyAddr());
  HCUpdate_Init(&update, action, &myentry, 0); 
  if(action == HC_InvalToParent){
    hcprop_handleInvalToParent(&update, 1);
  }
  else{
    assert(action == HC_InformToParent);
    hcprop_handleInformToParent(&update, 1);
  }
}



/*
 *------------------------------------------------------------------
 *
 * hcprop_handleInvalToChild --
 *
 *          Handle an invalidation message from a parent to me
 *          acting as a child. If I know
 *          of a different copy of the data, tell parent.
 *          Otherwise, tell children.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 * Note:
 *      Right now, we propagate message down tree even if we didn't
 *      know about the object being invalidated. Reasoning is that
 *      we might have discarded our hint due to limited associativity.
 *      But, the other reason (more common, probably) could be that
 *      I invalidated it, told my parent, and my parent told all
 *      of its children (including me.) When a child invalidates
 *      a parents copy, all descendents of the parent hear about
 *      it, which may not be right. I think we can live with
 *      this; essentially, when a parent's copy gets invalidated,
 *      all nodes in its subtree hear about it. If a node invalidates
 *      a copy which is shared by T nodes in a subtree, the node
 *      will hear echos of the invalidate logT times. Put another
 *      way, in a k-ary tree where the object is referenced up to
 *      level l, the best we could do would be send M = k^l messages.
 *      Under this scheme, we send M + M/k + M/k^2 + ... M/k^l
 *      messages, which for reasonable sized k is not too many extras
 *      (at worst, for binary tree, we send <2M messages.)
 *      
 *-------------------------------------------------4.2.1998----------
 */
void
hcprop_handleInvalToChild(HCUpdate *update)
{
  int stillHave;
  HCEntry otherCopy;
  int pIndex, cIndex;
  int iAmRoot, iAmLeaf;

  stillHave = HCDisk_netInvalRecord(hcDisk, &update->entry, &otherCopy);
  pIndex = HCHier_ParentQIndex(&update->entry.key, &iAmRoot);
  if(stillHave){
    /*
     * Msg from parent didn't invalidate the copy I know about.
     * Tell parent about it. 
     * Would be better to only tell parent if this copy is
     * in my subtree. But hierarchy can't tell me that.
     */
    if(!iAmRoot){
      assert(pIndex < HCHier_NParentQs());
      HCNet_Enqueue(&parentOutgoingA[pIndex], HC_InformToParent, 
		    &otherCopy, update->hopcount+1);
    }
  }
  else{
    cIndex = HCHier_ChildQIndex(&update->entry.key, iAmRoot, &iAmLeaf);
    if(!iAmLeaf){
      assert(cIndex < HCHier_NChildrenQs());
      HCNet_Enqueue(&childrenOutgoingA[cIndex], HC_InvalToChild, 
		    &update->entry, update->hopcount+1);
    }
  }
}


/*
 *------------------------------------------------------------------
 *
 * hcprop_handleInvalToParent --
 *
 *          Handle an invalidation from one of my children when
 *          I am acting as parent: if I know
 *          of a different copy of the data, squash message
 *          (see argument below). Otherwise, tell children
 *          and parent.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *-------------------------------------------------4.2.1998----------
 */
void
hcprop_handleInvalToParent(HCUpdate *update, int local)
{
  int stillHave;
  HCEntry otherCopy;
  int pIndex, cIndex;
  int iAmRoot, iAmLeaf;

  if(local){
    HCDisk_invalLocalCopy(hcDisk, update->entry.key);
    stillHave = 0;
  }
  else{
    stillHave = HCDisk_netInvalRecord(hcDisk, &update->entry, &otherCopy);
  }

  if(stillHave){
    /*
     * Invariant -- I know about 1 copy below me and that
     * is the only copy I tell my parents and children about.
     * Since this message didn't invalidate the one I knew
     * about, then this must NOT be the copy I told others
     * about. Squash message here.
     */
    ; /* do nothing -- squash message */
  }
  else{
    pIndex = HCHier_ParentQIndex(&update->entry.key, &iAmRoot);
    if(!iAmRoot){
      assert(pIndex < HCHier_NParentQs());
      HCNet_Enqueue(&parentOutgoingA[pIndex], HC_InvalToParent,
		    &update->entry, update->hopcount+1);
    }
    cIndex = HCHier_ChildQIndex(&update->entry.key, iAmRoot, &iAmLeaf);
    if(!iAmLeaf){
      assert(cIndex < HCHier_NChildrenQs());
      HCNet_Enqueue(&childrenOutgoingA[cIndex], HC_InvalToChild,
		    &update->entry, update->hopcount+1);
    }
  }
}




/*
 *------------------------------------------------------------------
 *
 * hcprop_handleInformToChild --
 *
 *          Handle a message from parent to me (acting as child)
 *          telling me about a new copy of an object. If this 
 *          copy is closer than the one I knew about before, 
 *          tell children about it.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *-------------------------------------------------4.2.1998-----------
 */
void
hcprop_handleInformToChild(HCUpdate *update)
{
  int didUpdate;
  int index;
  int iAmLeaf, iAmRoot;

  didUpdate = HCDisk_updateIfCloser(hcDisk, &update->entry);

  /*  printf("HCDataArrives at parent: 0x%x at <%s,%d>\n",
	 *(int *) &update->entry.key.key, inet_ntoa(update->entry.ipaddr),
	 update->entry.port); */

  if(didUpdate){
    (void)HCHier_ParentQIndex(&update->entry.key, &iAmRoot);
    index = HCHier_ChildQIndex(&update->entry.key, iAmRoot, &iAmLeaf);
    if(!iAmLeaf){
      assert(index < HCHier_NChildrenQs());
      HCNet_Enqueue(&childrenOutgoingA[index], HC_InformToChild,
		    &update->entry, update->hopcount+1);
    }
  }
}




/*
 *------------------------------------------------------------------
 *
 * hcprop_handleInformToParent --
 *
 *          Handle a message from child telling me (acting as parent)
 *          about a new copy of an object. If this copy is closer than 
 *          the one I knew about before, tell all children and
 *          parent.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------4.2.1998----------
 */
void
hcprop_handleInformToParent(HCUpdate *update, int local)
{
  int didUpdate;
  int pIndex, cIndex;
  int iAmRoot, iAmLeaf;

  if(local){
    HCDisk_informLocalCopy(hcDisk, update->entry.key);
    didUpdate = 1;
  }
  else{
    didUpdate = HCDisk_updateIfCloser(hcDisk, &update->entry);
  }
  /*  printf("HCDataArrives at parent: 0x%x at <%s,%d>\n",
	 *(int *) &update->entry.key.key, inet_ntoa(update->entry.ipaddr),
	 update->entry.port);*/
  if(didUpdate){
    pIndex = HCHier_ParentQIndex(&update->entry.key, &iAmRoot);
    if(!iAmRoot){
      assert(pIndex < HCHier_NParentQs());
      HCNet_Enqueue(&parentOutgoingA[pIndex], HC_InformToParent,
		    &update->entry, update->hopcount+1);
    }
    cIndex = HCHier_ChildQIndex(&update->entry.key, iAmRoot, &iAmLeaf);
    if(!iAmLeaf){
      assert(cIndex < HCHier_NChildrenQs());
      HCNet_Enqueue(&childrenOutgoingA[cIndex], HC_InformToChild,
		    &update->entry, update->hopcount+1);
    }
  }
}

