/*
 *------------------------------------------------------------------
 *
 * LESS Group 
 * Computer Sciences Department 
 * University of Texas at Austin
 *
 * HCDisk ---
 *          The on-disk cache of hints. The key data structure
 *          is an mmapped array which is treated as a <entriesPerBucket>-
 *          associative cache.
 *
 * $Date: 1998/02/01 21:14:52 $ $Id: HCDisk.c,v 1.1 1998/02/01 21:14:52 jkay Exp $
 *
 * Copyright (c) 1997 by Mike Dahlin.
 * 
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for noncommercial use without fee and
 * without written agreement is hereby granted, provided that 
 * the above copyright notice appear in all copies of this software.
 * 
 *------------------------------------------------------------------
 */
#include <assert.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <unistd.h>

#include "HCPrimitives.h"
#include "HCDisk.h"

static const char *suffix = ".theCache";

static void initializeFile(FILE *dataF, int nbuckets, int associativity);
static void init(HCDisk *d, char *diskPath);
static NodeId deleteKey(HCDisk *d, HCEntry *bucketp, URLKey key, NodeId node);

static NodeId findKey(HCDisk *d, HCEntry *bucketp, URLKey key);
static void  insertKey(HCDisk *d, HCEntry *bp, URLKey key, NodeId node);
static caddr_t pageAddr(caddr_t e);
static HCEntry *bucket(HCDisk *d, URLKey key);
static void timeMadvise();

static long pageSize;

/*
 * Should we try to keep stuff we just read or wrote in the cache.
 */
/*
 * You may not want to cache writes since they may or may not
 * be about stuff you will later read. On the other hand, there
 * may be temporal locality about stuff that gets updated --
 * I may be likely to read or discard the object in the near
 * future. Also, telling system to kick it out of cache costs
 * about 1ms per call.
 */
static const int CACHE_RECENT_WRITES = 1;
/*
 * You may not want to cache reads since these are things
 * that you will now have in your cache (so why would
 * you look at the hints any time soon. On the other hand,
 * telling system to kick out reads costs about 1ms per call
 */
static const int CACHE_RECENT_READS = 0;
/* 
 * You may not want to prefetch at all (we can only prefetch
 * on network updates, and those are not on critical
 * path anywhere. Maybe better to sacrifice latency for
 * reduced overhead.
 * 
 * No point int trying to pipeline if there are not many
 * updates to pipeline. Otherwise the overhead of the
 * system calls outweighs the reduced latency for disk
 * reads. In the best case, madvise (tight loop of madvise
 * about same location), madvise appears to take about 1ms per call.
 */
static const int DO_PREFETCH = 0;
static const int PREFETCH_THRESHOLD = 3;

/*
 *------------------------------------------------------------------
 *
 * HCDisk_Init --
 *
 *          Constructor
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
void 
HCDisk_Init(HCDisk *d, char *diskPath, HCHier *h)
{
  d->theHier = h;

  if((pageSize = sysconf(_SC_PAGESIZE)) < 0){
    perror("Cannot determine page size\n");
    exit(-1);
  }
  init(d, diskPath);  /* sets d->nbuckets, entriesPerBucket */
}

/*
 *------------------------------------------------------------------
 *
 * init --
 *
 *          Does teh set-up work
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
static void
init(HCDisk *d, char *diskPath)
{
  FILE *header;
  int dataF;
  const static int BIG_ENOUGH = 1024;
  char buffer[BIG_ENOUGH], token[BIG_ENOUGH], dataFile[BIG_ENOUGH];
  int version;
  int size, buckets, associativity;
  int gotVersion = 0, gotData = 0, gotAssoc = 0, 
    gotSize = 0, gotBuckets = 0;


  header = fopen(diskPath, "r");
  if(!header){
    perror("Opening Hint Cache header");
    exit(-1);
  }

  while(!feof(header)){
    if(!fgets(buffer, BIG_ENOUGH, header)){
      if(feof(header)){
	break;
      }
      else{
	perror("Reading hint cache header");
	exit(-1);
      }
    }
    if(buffer[0] != '#'){
      sscanf(buffer, "%s", token);
      if(!strcmp(token, "VERSION")){
	gotVersion = sscanf(buffer, "VERSION %d\n", &version);
	assert(gotVersion == 1);
	assert(version == 1);
      }
      else if(!strcmp(token, "DATA_FILE")){
	gotData = sscanf(buffer, "DATA_FILE %s\n", dataFile);
	assert(gotData == 1);
      }
      else if(!strcmp(token, "SIZE_BYTES")){
	gotSize = sscanf(buffer, "SIZE_BYTES %d\n", &size);
	assert(gotSize == 1);
      }
      else if(!strcmp(token, "BUCKETS")){
	gotBuckets = sscanf(buffer, "BUCKETS %d\n", &buckets);
	assert(gotBuckets == 1);
      }
      else if(!strcmp(token, "ASSOCIATIVITY")){
	gotAssoc = sscanf(buffer, "ASSOCIATIVITY %d\n", &associativity);
	assert(gotAssoc == 1);
      }
      else{
	/* Ignore what we don't understand */
      }
    }
  }
  fclose(header);
  assert(gotVersion && gotData && gotAssoc && gotSize && gotBuckets);
  assert(size == buckets * associativity * sizeof(HCEntry));
  d->nbuckets = buckets;
  d->entriesPerBucket = associativity;
  dataF =  open(dataFile, O_RDWR);
  if(dataF < 0){
    perror("Opening Hint Cache data");
    exit(-1);
  }
  d->mmappedArray = (HCEntry *)mmap(0, size, PROT_READ|PROT_WRITE,
				   MAP_SHARED, dataF, 0);
  if(!d->mmappedArray){
    perror("MMapping hint cache data");
    exit(-1);
  }
  if(madvise((char *)d->mmappedArray, size, MADV_RANDOM)){
    perror("WARNING: Madvise on hint cache data failed");
  }    
  close(dataF);
}


/*
 *------------------------------------------------------------------
 *
 * HCDisk_CreateFile --
 *
 *          Does the hard work of setting up a new empty file.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
void 
HCDisk_CreateFile(char *hcPath, int size, int associativity)
{
  char *dataPath;
  int buckets;
  FILE *header, *data;

  dataPath = (char *)malloc(strlen(hcPath) + strlen(suffix) + 1);
  assert(dataPath);
  sprintf(dataPath, "%s%s", hcPath, suffix);

  buckets = (size / associativity) / sizeof(HCEntry);
  assert(buckets * associativity * sizeof(HCEntry) <= size);
  size = buckets * associativity * sizeof(HCEntry);

  header = fopen(hcPath, "w");
  if(!header){
    perror("Opening Hint Cache header");
    exit(-1);
  }
  fprintf(header, "# *** Header for hint cache. Do not modify ***\n");
  fprintf(header, "VERSION 1\n");
  fprintf(header, "DATA_FILE %s\n", dataPath);
  fprintf(header, "SIZE_BYTES %d\n", size);
  fprintf(header, "BUCKETS %d\n", buckets);
  fprintf(header, "ASSOCIATIVITY %d\n", associativity);
  fclose(header);
  
  data = fopen(dataPath, "w");
  if(!data){
    perror("Opening Hint Cache data");
    exit(-1);
  }
  initializeFile(data, buckets, associativity);
  fclose(data);
  free(dataPath);
}

/*
 *------------------------------------------------------------------
 *
 * initializeFile --
 *
 *          Set a file to empty by invalidating all entries.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
static void
initializeFile(FILE *dataF, int nbuckets, int associativity)
{
  int ibucket, ientry;
  int count;

  HCEntry invalidEntry;
  HCEntry_Init(&invalidEntry, INVALID_URL_KEY, INVALID_NODE_ID);

  for(ibucket = 0; ibucket < nbuckets; ibucket++){
    for(ientry = 0; ientry < associativity; ientry++){
      count = fwrite(&invalidEntry, sizeof(HCEntry), 1, dataF);
      if(count != 1){
	perror("ERROR Creating HCCache file");
	exit(-1);
      }
    }
  }
}

/*
 *------------------------------------------------------------------
 *
 * HCDisk_Close --
 *
 *          Unmap the file.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
void HCDisk_Close(HCDisk *d)
{
  munmap((char *)d->mmappedArray, d->nbuckets * d->entriesPerBucket);
  d->mmappedArray = NULL;
}

/*
 *------------------------------------------------------------------
 *
 * HCDIsk_informLocalCopy --
 *
 *          Update data structures to reflect the fact a copy is
 *          cached locally.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
void HCDisk_informLocalCopy(HCDisk *d, URLKey key, NodeId node)
{
  HCEntry *bp;

  assert(d);

  if(!URLKey_Compare(&key, &INVALID_URL_KEY) || !NodeId_valid(&node)){
    return;
  }

  bp = bucket(d, key);
  deleteKey(d, bp, key, INVALID_NODE_ID);
  insertKey(d, bp, key, node);
  if(!CACHE_RECENT_WRITES){
    if(madvise(pageAddr((caddr_t)bp), pageSize, MADV_DONTNEED)){
      perror("Madvise DONTNEED failed\n");
    }
  }
}



/*
 *------------------------------------------------------------------
 *
 * HCDisk_InvalLocalCopy --
 *
 *          Update local data structures to reflect the fact
 *          that we no longer have a copy locally.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
void HCDisk_invalLocalCopy(HCDisk *d, URLKey key, NodeId myId)
{
  HCEntry *bp;

  assert(d);

  if(!URLKey_Compare(&key, &INVALID_URL_KEY) || !NodeId_valid(&myId)){
    return;
  }

  bp = bucket(d, key);
  deleteKey(d, bp, key, myId);
  if(madvise(pageAddr((caddr_t)bp), pageSize, MADV_DONTNEED)){
    perror("Madvise DONTNEED failed\n");
  }
}

/*
 *------------------------------------------------------------------
 *
 * HCPrefetch --
 *
 *          description.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
void HCDisk_prefetch(HCDisk *d, HCUpdate *u, int nupdates)
{
  int ii;
  HCEntry *bp;

  assert(d);
  assert(u);

  if(DO_PREFETCH && (nupdates >= PREFETCH_THRESHOLD)){
    for(ii = 0; ii < nupdates; ii++){
      bp = bucket(d, u[ii].entry.key);
      if(madvise(pageAddr((caddr_t)bp), pageSize, MADV_WILLNEED)){
	perror("Madvise WILLNEED failed\n");
      }
    }
  }
}


/*
 *------------------------------------------------------------------
 *
 * HCDisk_netInvalRecord --
 *
 *          The network told us that <key, invalNode> is
 *          no longer a valid mapping.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
NodeId 
HCDisk_netInvalRecord(HCDisk *d, URLKey key, NodeId invalNode)
{
  HCEntry *bp;
  NodeId ret;

  assert(d);

  if(!URLKey_Compare(&key, &INVALID_URL_KEY) || !NodeId_valid(&invalNode)){
    return INVALID_NODE_ID;
  }

  bp = bucket(d, key);
  ret = deleteKey(d, bp, key, invalNode);
  if(madvise(pageAddr((caddr_t)bp), pageSize, MADV_DONTNEED)){
    perror("Madvise DONTNEED failed\n");
  }
  return ret;
}

/*
 *------------------------------------------------------------------
 *
 * HCDisk_updateIfCloser --
 *
 *          The network told us that <key, newnode> is a valid
 *          mapping. Install it if it is better than the one
 *          we have.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
int HCDisk_updateIfCloser(HCDisk *d, URLKey key, NodeId newNode)
{
  HCEntry *bp;
  NodeId oldNode;
  int ret = 0;
  assert(d);

  if(!URLKey_Compare(&key, &INVALID_URL_KEY) || !NodeId_valid(&newNode)){
    return 0;
  }

  bp = bucket(d, key);
  oldNode = findKey(d, bp, key);
  if(HCHier_compareDistanceFromMe(d->theHier, newNode, oldNode) < 0){
    deleteKey(d, bp, key, oldNode);
    insertKey(d, bp, key, newNode);
    ret = 1;
  }
  if(!CACHE_RECENT_WRITES){
    if(madvise(pageAddr((caddr_t)bp), pageSize, MADV_DONTNEED)){
      perror("Madvise DONTNEED failed\n");
    }
  }
  return ret;
}


NodeId 
HCDisk_findNearest(HCDisk *d, URLKey key)
{
  HCEntry *bp;
  NodeId node;

  assert(d);

  if(!URLKey_Compare(&key, &INVALID_URL_KEY)){
    return INVALID_NODE_ID;
  }

  bp = bucket(d, key);
  node = findKey(d, bp, key);
  if(!CACHE_RECENT_READS){
    if(madvise(pageAddr((caddr_t)bp), pageSize, MADV_DONTNEED)){
      perror("Madvise DONTNEED failed\n");
    }
  }
  return node;
}


/*
 *------------------------------------------------------------------
 *
 * bucket --
 *
 *          Return pointer to first entry in bucket for key.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
static HCEntry *
bucket(HCDisk *d, URLKey key)
{
  int b;

  assert(d);
  assert(d->mmappedArray);

  b = key.key % d->nbuckets;
  return &(d->mmappedArray[b * d->entriesPerBucket]);
}

/*
 *------------------------------------------------------------------
 *
 * pageAddr --
 *
 *          Return the address of the first byte of the
 *          page that holds an address.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
static caddr_t
pageAddr(caddr_t e)
{
  caddr_t ret;
  ret = e - ((unsigned)e % pageSize);
  return ret;
}

/*
 *------------------------------------------------------------------
 *
 * deleteKey --
 *
 *          Delete the record associated with the specified key, node
 *          pair, or with the specified key (and any node) if the
 *          node passed in is NULL. If there is a record for the key that
 *          remains after the delete, return its nodeID.
 *          
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
static NodeId
deleteKey(HCDisk *d, HCEntry *bucketp, URLKey key, NodeId node)
{
  int ii, gotIt;
  NodeId returnNoNode;

  gotIt = 0;
  assert(URLKey_Compare(&key, &INVALID_URL_KEY));

  for(ii = 0; ii < d->entriesPerBucket; ii++){
    if(gotIt){
      /*
       * Removed earlier entry -- shift towards 0
       */
      assert(ii > 0);
      HCEntry_Copy(&bucketp[ii-1], &bucketp[ii]);
    }
    else{
      if(!URLKey_Compare(&bucketp[ii].key, &key)){
	if(!NodeId_Compare(&node, &INVALID_NODE_ID) 
	   || !NodeId_Compare(&node, &bucketp[ii].node)){
	  gotIt = 1;
	  bucketp[ii].key = INVALID_URL_KEY;
	  NodeId_markInvalid(&bucketp[ii].node);
	}
	else{
	  /*
	   * Matched key but not NodeId -- return
	   * node we know about that we thought was better than
	   * the thing you were invalidating
	   */
	  return bucketp[ii].node;
	}
      }
    }
  }
  NodeId_markInvalid(&returnNoNode);
  return returnNoNode;
}


/*
 *------------------------------------------------------------------
 *
 * insertKey --
 *
 *          description.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
static void  insertKey(HCDisk *d, HCEntry *bucketp, URLKey key, NodeId node)
{
  int ii;

  assert(URLKey_Compare(&key, &INVALID_URL_KEY));
  assert(NodeId_valid(&node));

  /*
   * Find empty entry. 
   */
  for(ii = 0; ii < d->entriesPerBucket; ii++){
    if(!NodeId_valid(&bucketp[ii].node)){
      break;
    }
  }
  /*
   * No empty entry -- make one by shifting towards 0
   */
  if(ii == d->entriesPerBucket){
    for(ii = 1; ii <  d->entriesPerBucket; ii++){
      HCEntry_Copy(&bucketp[ii-1], &bucketp[ii]);
    }
    ii = d->entriesPerBucket - 1;
  }
  bucketp[ii].key = key;
  bucketp[ii].node = node; 
}

/*
 *------------------------------------------------------------------
 *
 * findKey --
 *
 *          Find an entry that matches the key and return
 *          the node for that key. Since this is a read, 
 *          update the LRU list by removing key and
 *          then re-inserting key.
 *         
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
static NodeId
findKey(HCDisk *d, HCEntry *bucketp, URLKey key)
{
  int ii;
  NodeId location;

  assert(URLKey_Compare(&key, &INVALID_URL_KEY));
  for(ii = 0; ii < d->entriesPerBucket; ii++){
    if(!URLKey_Compare(&bucketp[ii].key, &key)){
      location = bucketp[ii].node;
      deleteKey(d, bucketp, key, location);
      insertKey(d, bucketp, key, location);
      return location;
    }
  }
  NodeId_markInvalid(&location);
  return location;
}



/*
 *------------------------------------------------------------------
 *
 * HCDisk_SelfTest --
 *
 *          description.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
void
HCDisk_selfTest()
{
  HCDisk d;
  char *testPath = "DiskSelfTestFile";
  char *testHier = "testHierarchyFile";
  char urlString[1024];
  URLKey url, oldUrl, olderUrl;
  NodeId n0, n1, n2, n3, n4, n5, n6, place;
  HCHier h;
  int size, assoc, ii;
  int maxSize, maxAssoc, maxi;
  const static int doBigTest = 0;
  struct sockaddr *s;

  if(doBigTest){
    maxSize = 10000;
    maxAssoc = 8;
    maxi = 100000;
  }
  else{
    maxSize = 1000;
    maxAssoc = 5;
    maxi = 100;
  }


  printf("HCDisk_selfTest..."); fflush(stdout);

  timeMadvise();

  NodeId_Init(&n0, 0x80536F03, 0);
  NodeId_Init(&n1, 0x80536F03, 1);
  NodeId_Init(&n2, 0x80536F03, 2);
  NodeId_Init(&n3, 0x80536F03, 3);
  NodeId_Init(&n4, 0x80536F03, 4);
  NodeId_Init(&n5, 0x80536F03, 5);
  NodeId_Init(&n6, 0x80536F03, 6);

  s = NodeId_copyAsSockAddr(&n5);  /* I am node 5*/
  HCHier_Init(&h, testHier, s);

  for(size = 8 * sizeof(HCEntry); size < maxSize; size *= 10){ 
    for(assoc = 1; assoc < maxAssoc; assoc+=3){
      printf("."); fflush(stdout);
      HCDisk_CreateFile(testPath, size, assoc);
      HCDisk_Init(&d, testPath, &h);

      url = INVALID_URL_KEY;
      oldUrl = INVALID_URL_KEY;
      olderUrl = INVALID_URL_KEY;

      for(ii = 0; ii < maxi; ii++){
	oldUrl = url;
	sprintf(urlString, "http://wlksjdf%ddlkfjskdf", ii);
	URLKey_Init(&url, urlString);
	/*
	 * Check miss 
	 */
	place = HCDisk_findNearest(&d, url);
	assert(!NodeId_Compare(&place, &INVALID_NODE_ID));
	/*
	 * Write then read
	 */
	assert(HCDisk_updateIfCloser(&d, url, n1));
	place = HCDisk_findNearest(&d, url);
	assert(!NodeId_Compare(&place, &n1));
	/*
	 * check net overwrite with closer
	 */
	assert(HCDisk_updateIfCloser(&d, url, n6));
	place = HCDisk_findNearest(&d, url);
	assert(!NodeId_Compare(&place, &n6));
	/*
	 * Check net no overwrite with farther
	 */
	assert(!HCDisk_updateIfCloser(&d, url, n3));
	place = HCDisk_findNearest(&d, url);
	assert(NodeId_Compare(&place, &n3));
	assert(!NodeId_Compare(&place, &n6));
	/*
	 * Check net invalidate -- different copy
	 */
	place = HCDisk_netInvalRecord(&d, url, n1);
	assert(!NodeId_Compare(&place, &n6));
	place = HCDisk_findNearest(&d, url);
	assert(!NodeId_Compare(&place, &n6));
	/*
	 * Check net invalidate -- of the one we've got
	 */
	place = HCDisk_netInvalRecord(&d, url, n6);
	assert(!NodeId_Compare(&place, &INVALID_NODE_ID));
	place = HCDisk_findNearest(&d, url);
	assert(!NodeId_Compare(&place, &INVALID_NODE_ID));
	/*
	 * Check local update
	 */
	HCDisk_informLocalCopy(&d, url, n5);
	place = HCDisk_findNearest(&d, url);
	assert(!NodeId_Compare(&place, &n5));
	/*
	 * Check local invalidate
	 */
	HCDisk_invalLocalCopy(&d, url, n5);
	place = HCDisk_findNearest(&d, url);
	assert(!NodeId_Compare(&place, &INVALID_NODE_ID));


	/* 
	 * Leave local copy when done
	 */
	HCDisk_informLocalCopy(&d, url, n5);
	place = HCDisk_findNearest(&d, url);
	assert(!NodeId_Compare(&place, &n5));

	/*
	 * check associativity by reading from past passes
	 */
	if(assoc > 1){
	  if(URLKey_Compare(&oldUrl, &INVALID_URL_KEY)){
	    place = HCDisk_findNearest(&d, oldUrl);
	    assert(!NodeId_Compare(&place, &n5));
	  }
	}
	if(assoc > 2){
	  if(URLKey_Compare(&olderUrl, &INVALID_URL_KEY)){
	    place = HCDisk_findNearest(&d, olderUrl);
	    assert(!NodeId_Compare(&place, &n5));
	  }
	}

      }  
      HCDisk_Close(&d);
      remove(testPath);
    }
  }
  free(s);
  printf("..OK\n");
}


static void
timeMadvise()
{
  struct timeval start, end;
  char *data, *p;
  int ii;
  float time;
  static const int count = 10000;

  if((pageSize = sysconf(_SC_PAGESIZE)) < 0){
    perror("Cannot determine page size\n");
    exit(-1);
  }

  data = (char *)malloc(16382);
  p = pageAddr(data);

  assert(!gettimeofday(&start, NULL));
  for(ii = 0; ii < count; ii++){
    madvise(p, pageSize, MADV_DONTNEED);
  }
  assert(!gettimeofday(&end, NULL));
  time =  end.tv_sec - start.tv_sec;
  time = time + (float)(end.tv_usec - start.tv_usec)/(float)(1000000.0);

  printf("Called Madvise %d times in %f seconds: %f per call\n",
	 count, time, time/(float)count);
  free(data);
}
