/*
 *------------------------------------------------------------------
 *
 * LESS Group 
 * Computer Sciences Department 
 * University of Texas at Austin
 *
 * HCHintCache.h ---
 *          High-level interface to hint cache.
 *
 * $Date: 1998/02/01 21:14:52 $ $Id: HCHintCache.c,v 1.1 1998/02/01 21:14:52 jkay Exp $
 *
 * Copyright (c) 1997 by Mike Dahlin.
 * 
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for noncommercial use without fee and
 * without written agreement is hereby granted, provided that 
 * the above copyright notice appear in all copies of this software.
 * 
 *------------------------------------------------------------------
 */
#include <assert.h>

#include "squid.h"
#include "HCHintCache.h"
#include "HCInternals.h"
#include "HCPrimitives.h"

#define HC_TIMEOUT  30



static void localAction(char *url, enum HCUpdateEnum action);

/*
 *------------------------------------------------------------------
 *
 * HCInit --
 *
 *          description.
 *
 * Arguments:
 *      char *hcPath -- config file for cache
 *      char *hierPath -- config file for hierarchy
 *      sockaddr *myAddr -- my IP address
 *
 * Results:
 *      Return nonzero on error.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
int
HCInit(char *hcPath, char *hierPath, struct sockaddr *myAddr)
{
/* 
 *XXX I assume that the struct sockaddr * is a pointer to a thing
 *XXX that will never go away This is wrong. Make a copy.
 */
  assert(0);


  /* Start hint timer */
  eventAdd("hint timer", (EVH) HCTimer, NULL, HC_TIMEOUT);

  hcHier = (HCHier *)malloc(sizeof(HCHier));
  assert(hcHier);
  if(HCHier_Init(hcHier, hierPath, myAddr)){
    return -1;
  }


  hcDisk = (HCDisk *)malloc(sizeof(HCDisk));
  assert(hcDisk);
  HCDisk_Init(hcDisk, hcPath, hcHier);
  
  childrenOutgoing = (HCNet *)malloc(sizeof(HCNet));
  assert(childrenOutgoing);
  HCNet_Init(childrenOutgoing);
  parentOutgoing = (HCNet *)malloc(sizeof(HCNet));
  assert(parentOutgoing);
  HCNet_Init(parentOutgoing);

  return 0;
}

/*
 *------------------------------------------------------------------
 *
 * HCDestroy --
 *
 *          Shut down.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
void
HCDestroy()
{
  HCHier_Destroy(hcHier);
  free(hcHier);
  HCDisk_Close(hcDisk);
  free(hcDisk);
  hcDisk = NULL;
  HCNet_Destroy(childrenOutgoing);
  free(childrenOutgoing);
  childrenOutgoing = NULL;
  HCNet_Destroy(parentOutgoing);
  free(parentOutgoing);
  parentOutgoing = NULL;
}

/*
 *------------------------------------------------------------------
 *
 * HCCreate --
 *
 *          Create a NEW disk cache (obliterating the one
 *          already on disk, if any). Then do normal initialization.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
void 
HCCreate(char *hcPath, int size, char *hierPath, struct sockaddr *myAddr)
{
  const static int defaultAssociativity = 4;

  HCDisk_CreateFile(hcPath, size, defaultAssociativity);
  HCInit(hcPath, hierPath, myAddr);
}

/*
 *------------------------------------------------------------------
 *
 * HCinformLocalCopy --
 *
 *          description.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
void 
HCinformLocalCopy(char *url)
{
  localAction(url, HC_ChildInform);
}
 
void 
HCinvalLocalCopy(char *url)
{
  localAction(url, HC_ChildInval);
}

static void
localAction(char *url, enum HCUpdateEnum action)
{
  URLKey key;
  NodeId myId;

  assert(action == HC_ChildInform || action == HC_ChildInval);

  URLKey_Init(&key, url);
  myId = HCHier_myNodeId(hcHier);

  /*
   * Put it on the network
   */
  HCNet_Enqueue(parentOutgoing, action, key, myId);

  /*
   * Update local disk data structures.
   */
  if(action == HC_ChildInform){
    HCDisk_informLocalCopy(hcDisk, key, myId);
  }
  else{
    assert(action == HC_ChildInval);
    HCDisk_invalLocalCopy(hcDisk, key, myId);
  }
}


/*
 * Warning -- this returns a newly allocated sockaddr; If this is bad, we
 * can change NodeId to have a sockaddr structure internally and
 * return pointers to that thing. Need to be careful in that case
 * about returning pointers to internal parts of automatic variables
 */
struct sockaddr *
HCfindNearest(char *url)
{
  URLKey urlKey;
  struct sockaddr *copy;
  NodeId nodeId;

  URLKey_Init(&urlKey, url);
  nodeId = HCDisk_findNearest(hcDisk, urlKey);
  copy = NodeId_copyAsSockAddr(&nodeId);
  printf("XXX1: Hey, john do you free() the sockaddr* that I give you in HCFindNearest()?\n");
  return copy;
}


/*
 *------------------------------------------------------------------
 *
 * HCTimer --
 *
 *          This gets called every so often. Its job is
 *          to flush outgoing messages.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
void HCTimer(void)
{
  NodeId destN;
  int ii;

  /* Reload timer */
  eventAdd("hint timer", (EVH) HCTimer, NULL, HC_TIMEOUT);

  if(HCNet_bytesReady(childrenOutgoing) > 0){
    HCNet_Complete(childrenOutgoing);
    for(ii = 0; ii < HCHier_nChildren(hcHier); ii++){
      destN = HCHier_getChild(hcHier, ii);
      HCNet_SendTo(childrenOutgoing, destN);
    }
    HCNet_Done(childrenOutgoing);
  }

  if(HCNet_bytesReady(parentOutgoing) > 0){
    HCNet_Complete(parentOutgoing);
    /*
     * For multiple parents, will need to enqueu in multiple
     * send queues at send time. E.g. parentOutgoing will
     * become parentOutgoing[].
     */
    assert(HCHier_nParents(hcHier) == 1);
    destN = HCHier_getParent(hcHier, 0);
    HCNet_SendTo(parentOutgoing, destN);
    HCNet_Done(parentOutgoing);
  }
}


void
HCDataArrives(char *data, int len, struct sockaddr *source)
{
  NodeId node, nearestCopy;
  Net_HCUpdate *netUpdates;
  int nUpdates, iupdate;
  HCUpdate *updates;
  int didUpdate;

  NodeId_copyFromSockAddr(&node, source);
  netUpdates = (Net_HCUpdate *)data;
  nUpdates = len / sizeof(Net_HCUpdate);
  if(len % sizeof(Net_HCUpdate)){
    fprintf(stderr, 
	    "HCHintCache: warning -- unexpected packet size from net\n");
  }
  updates = (HCUpdate *)malloc(nUpdates * sizeof(HCUpdate));
  if(!updates){
    /*
     * On error, drop updates on floor. They're just hints.
     */
    return;
  }
  for(iupdate = 0; iupdate < nUpdates; iupdate++){
    Net_HCUpdate_LocalFormat(&netUpdates[iupdate], &updates[iupdate]);
  }
  HCDisk_prefetch(hcDisk, updates, nUpdates);

  for(iupdate = 0; iupdate < nUpdates; iupdate++){
    switch(updates[iupdate].action){
    case HC_ParentInval:
      nearestCopy = HCDisk_netInvalRecord(hcDisk, 
					  updates[iupdate].entry.key,
					  updates[iupdate].entry.node);
      if(NodeId_valid(&nearestCopy) 
	 && HCHier_inMySubtree(hcHier, nearestCopy)){
	/*
	 * Msg from parent didn't invalidate the copy I know about.
	 * Tell parent about it.
	 */
	HCNet_Enqueue(parentOutgoing, HC_ChildInform,
		      updates[iupdate].entry.key, nearestCopy);
      }
      else{
	HCNet_Enqueue(childrenOutgoing, HC_ParentInval, 
		      updates[iupdate].entry.key, updates[iupdate].entry.node);
      }
      break;
    case HC_ChildInval:
      nearestCopy = HCDisk_netInvalRecord(hcDisk, 
					  updates[iupdate].entry.key,
					  updates[iupdate].entry.node);
      if(NodeId_valid(&nearestCopy)){
	/*
	 * Invariant -- I know about 1 copy below me and that
	 * is the only copy I tell my parents and children about.
	 * Since this message didn't invalidate the one I knew
	 * about, then this must NOT be the copy I told others
	 * about. Squash message here. 
	 */
	; /* do nothing -- squash message */
      }
      else{
	HCNet_Enqueue(childrenOutgoing, HC_ParentInval, 
		      updates[iupdate].entry.key, updates[iupdate].entry.node);
	HCNet_Enqueue(parentOutgoing, HC_ChildInval, 
		      updates[iupdate].entry.key, updates[iupdate].entry.node);
      }
      break;
    case HC_ParentInform:
      didUpdate = HCDisk_updateIfCloser(hcDisk, updates[iupdate].entry.key, 
					updates[iupdate].entry.node); 
      if(didUpdate){
	HCNet_Enqueue(childrenOutgoing, HC_ParentInform, 
		      updates[iupdate].entry.key, updates[iupdate].entry.node);
      }
      break;
    case HC_ChildInform:
      didUpdate = HCDisk_updateIfCloser(hcDisk, updates[iupdate].entry.key, 
					updates[iupdate].entry.node); 
      if(didUpdate){
	HCNet_Enqueue(childrenOutgoing, HC_ParentInform, 
		      updates[iupdate].entry.key, updates[iupdate].entry.node);
	HCNet_Enqueue(parentOutgoing, HC_ChildInform, 
		      updates[iupdate].entry.key, updates[iupdate].entry.node);
      }
      break;
    default:
      /* Ignore bad messages */
      fprintf(stderr, 
	      "HCHintCache: warning -- unknown message action from net\n");
      break;
    }
  }
  free(updates);
}






/*
 *------------------------------------------------------------------
 *
 * HCHintCache_selfTest --
 *
 *          description.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
void
HCHintCache_selfTest()
{
  char *testPath = "DiskSelfTestFile";
  char *testHier = "testHierarchyFile";
  struct sockaddr *s;
  NodeId n0, n1, n2, n3, n4, n5, n6;
  struct sockaddr_in *ret, *s3;
  Net_HCUpdate netData[2];
  HCUpdate data[2];
  URLKey key[2];
  HCEntry entry[2];
  char name1[100], name2[100], name3[100];

  printf("HCHintCache selftest()\n");
  NodeId_Init(&n0, 0x80536F03, 0);
  NodeId_Init(&n1, 0x80536F03, 1);
  NodeId_Init(&n2, 0x80536F03, 2);
  NodeId_Init(&n3, 0x80536F03, 3);
  NodeId_Init(&n4, 0x80536F03, 4);
  NodeId_Init(&n5, 0x80536F03, 5);
  NodeId_Init(&n6, 0x80536F03, 6);
  sprintf(name1, "thing1");
  sprintf(name2, "thing2");
  sprintf(name3, "thing3");



  s = NodeId_copyAsSockAddr(&n5);  /* I am node 5*/
  HCCreate(testPath, 1024, testHier, s);


  HCinformLocalCopy(name1);
  ret = (struct sockaddr_in*)HCfindNearest(name1);
  assert(ret->sin_port == 5); /* hit */
  ret = (struct sockaddr_in*)HCfindNearest(name2);
  assert(ret->sin_port == 0);  /* miss */

  HCinvalLocalCopy(name1);
  ret = (struct sockaddr_in*)HCfindNearest(name1);
  assert(ret->sin_port == 0);  /* miss */

  HCTimer();

  s3 = (struct sockaddr_in*)NodeId_copyAsSockAddr(&n3);
  URLKey_Init(&key[0], name2);
  HCEntry_Init(&entry[0], key[0], n2);
  HCUpdate_Init(&data[0], HC_ParentInform, &entry[0]);
  Net_HCUpdate_Init(&netData[0], &data[0]);
  URLKey_Init(&key[1], name3);
  HCEntry_Init(&entry[1], key[1], n3);
  HCUpdate_Init(&data[1], HC_ParentInform, &entry[1]);
  Net_HCUpdate_Init(&netData[1], &data[1]);
  HCDataArrives((char *)netData, 2 * sizeof(Net_HCUpdate), (struct sockaddr *)s3);

  ret = (struct sockaddr_in*)HCfindNearest(name2);
  assert(ret->sin_port == 2);  /* hit at 2 */
  ret = (struct sockaddr_in*)HCfindNearest(name3);
  assert(ret->sin_port == 3);  /* hit at 3 */

  URLKey_Init(&key[0], name2);
  HCEntry_Init(&entry[0], key[0], n2);
  HCUpdate_Init(&data[0], HC_ParentInval, &entry[0]);
  Net_HCUpdate_Init(&netData[0], &data[0]);
  URLKey_Init(&key[1], name3);
  HCEntry_Init(&entry[1], key[1], n2);
  HCUpdate_Init(&data[1], HC_ParentInform, &entry[1]);
  Net_HCUpdate_Init(&netData[1], &data[1]);
  HCDataArrives((char *)netData, 2 * sizeof(Net_HCUpdate), (struct sockaddr *)s3);

  ret = (struct sockaddr_in*)HCfindNearest(name2);
  assert(ret->sin_port == 0);  /* miss-invalidated */
  ret = (struct sockaddr_in*)HCfindNearest(name3);
  assert(ret->sin_port == 2);  /* 2 is better than 3 so overwrite old */

  s3 = (struct sockaddr_in*)NodeId_copyAsSockAddr(&n3);
  URLKey_Init(&key[0], name2);
  HCEntry_Init(&entry[0], key[0], n2);
  HCUpdate_Init(&data[0], HC_ParentInform, &entry[0]);
  Net_HCUpdate_Init(&netData[0], &data[0]);
  URLKey_Init(&key[1], name3);
  HCEntry_Init(&entry[1], key[1], n3);
  HCUpdate_Init(&data[1], HC_ParentInform, &entry[1]);
  Net_HCUpdate_Init(&netData[1], &data[1]);
  HCDataArrives((char *)netData, 2 * sizeof(Net_HCUpdate), (struct sockaddr*)s3);

  ret = (struct sockaddr_in*)HCfindNearest(name2);
  assert(ret->sin_port == 2);  /* hit at 2 */
  ret = (struct sockaddr_in*)HCfindNearest(name3);
  assert(ret->sin_port == 2);  /* 2 is still better than 3 so no overwrite */

  HCTimer();


  HCDestroy();
  free(s3);
  free(s);
  printf("..OK\n");

}
