/*
 *------------------------------------------------------------------
 *
 * LESS Group 
 * Computer Sciences Department 
 * University of Texas at Austin
 *
 * HCPrimitives.h ---
 *          Primitive types for hint cache.
 *
 * $Date: 1998/02/01 21:14:52 $ $Id: HCPrimitives.c,v 1.1 1998/02/01 21:14:52 jkay Exp $
 *
 * Copyright (c) 1997 by Mike Dahlin.
 * 
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for noncommercial use without fee and
 * without written agreement is hereby granted, provided that 
 * the above copyright notice appear in all copies of this software.
 * 
 *------------------------------------------------------------------
 */
#include <assert.h>
#include <ctype.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <stdlib.h>
#include <stdio.h>

#include "HCEndian.h"
#include "HCPrimitives.h"
#include "md5.h"

const NodeId INVALID_NODE_ID = {0, 0};
const Net_NodeId INVALID_NET_NODE_ID = {0, 0};
const URLKey INVALID_URL_KEY = {0};

static long long unsigned calculateStupidKey(char *url);
static long long unsigned calculateMD5Key(char *url);
static void makeLowerCase(char *url);
static long long unsigned calculateKey(char *url);
static char toLower(char c);

/*
 *------------------------------------------------------------------
 *
 * NodeId --
 *
 *          A node IP address and port.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
void
NodeId_Init(NodeId *id, unsigned ipAddr_, u_short port_)
{
  assert(id);
  id->ipAddr = ipAddr_;
  id->port = port_;
}

int
NodeId_Compare(const NodeId *id1, const NodeId *id2)
{
  assert(id1);
  assert(id2);
  if(id1->ipAddr != id2->ipAddr){
    return 1;
  }
  if(id1->port != id2->port){
    return 1;
  }
  return 0;
}

void
NodeId_selfTest()
{
  assert(sizeof(unsigned) == sizeof(struct in_addr));
  /* assert(sizeof(u_short) == sizeof(struct sockaddr_in.sin_port)); */
}

void 
NodeId_copyFromSockAddr(NodeId *id, struct sockaddr *addr)
{
  struct sockaddr_in *internetAddr;
  
  if(addr->sa_family != PF_INET){
    assert(0);
  }
  internetAddr = (struct sockaddr_in*)addr;
  NodeId_Init(id, internetAddr->sin_addr.s_addr, internetAddr->sin_port);
}

/*
 * Returns a newly allocated sockaddr of subtype struct sockaddr_in
 * XXX Caller must free this pointer when done.
 */
struct sockaddr *
NodeId_copyAsSockAddr(NodeId *n)
{
  struct sockaddr_in *internetAddr;
  internetAddr = (struct sockaddr_in *)malloc(sizeof(struct sockaddr_in));
  assert(internetAddr);
  internetAddr->sin_family = PF_INET;
  internetAddr->sin_port = n->port;
  internetAddr->sin_addr.s_addr = n->ipAddr;
  return (struct sockaddr *)internetAddr;
}


void
Net_NodeId_Init(Net_NodeId *nnid, NodeId *nid)
{
  assert(nnid);
  nnid->mungedIpAddr = HCEndian_lMachineToNet(nid->ipAddr);
  nnid->mungedPort = HCEndian_sMachineToNet(nid->port);
}

void
Net_NodeId_LocalFormat(Net_NodeId *nnid, NodeId *nid)
{
  assert(nid);
  assert(nnid);
  nid->ipAddr = HCEndian_lNetToMachine(nnid->mungedIpAddr);
  nid->port = HCEndian_sNetToMachine(nnid->mungedPort);
}

int NodeId_valid(NodeId *id)
{
  if(!id){
    return 0;
  }
  return (NodeId_Compare(id, &INVALID_NODE_ID));
}

void NodeId_markInvalid(NodeId *id)
{
  if(!id){
    return;
  }
  *id = INVALID_NODE_ID;
}

/*
 * Suck a nodeID out of the buffer and return a pointer to
 * first character after nodeID or NULL if data doesn't
 * make sense as nodeID. Return the nodeID in result.
 * format is: 
 *     XXX.XXX.XXX.XXX:YYY
 * Where the XXX's are any integer value from 0..255 and YYY is
 * any integer value from 0..32768
 */
char *
NodeId_ParseString(NodeId *result, char *buf)
{
  unsigned a1, a2, a3, a4, p;
  int got;

  while(isspace(buf[0])){
    buf++;
  }
  got = sscanf(buf, "%u.%u.%u.%u:%u", &a1, &a2, &a3, &a4, &p);
  if((got != 5) || (a1 > 255) || (a2 > 255) 
     || (a3 > 255) || (a4 > 255) || (p > 32768)){
    *result = INVALID_NODE_ID;
    return NULL;
  }
  result->port = p;
  result->ipAddr = (a1 << 24) | (a2 << 16) | (a3 << 8) | a4;

  while(buf[0] != ':'){
    buf++;
  }
  while(!isspace(buf[0])){
    buf++;
  }
  return buf;
}



/*
 *------------------------------------------------------------------
 *
 * URLKey --
 *
 *          Constructor. Calculate a hash of the URL.
 *          This hash needs to be a good one; collisions
 *          must be rare because they get treated as misses
 *          or worse yet false hits.
 *
 *          Be sure to convert the URL to a canonical form.
 *          We assume that the only canonicalizing that
 *          we need to do is making it all the same case.
 *
 * Arguments:
 *      char *url -- a URL.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */

void
URLKey_Init(URLKey *key, char *url)
{
  assert(key);
  if(!url){
    key->key = 0;
  }

  makeLowerCase(url);
  key->key = calculateKey(url);    
}

int
URLKey_Compare(const URLKey *k1, const URLKey *k2)
{
  assert(k1);
  assert(k2);
  if(k1->key != k2->key){
    return 1;
  }
  return 0;
}


/*
 * Make lower case, but stop at third '/' (e.g., after the hostname).
 */
static void
makeLowerCase(char *url)
{
  int len, ii;
  int slashCount = 0;

  assert(url);
  len = strlen(url);
  for(ii = 0; ii < len; ii++){
    /*
     * Alignment.
     */
    if(url[ii] == '/'){
      slashCount++;
      if(slashCount == 3){
	return;
      }
    }
    url[ii] = toLower(url[ii]);
  }
}

static
char toLower(char c)
{
  static const char distance = 'A' - 'a';
  if(c >= 'A' && c <= 'Z'){
    c = c - distance;
    assert(c >= 'a' && c <= 'z');
  }
  return c;
}


/*
 * The hash function should be really good since we are counting
 * on there being few collisions in our k-way associative
 * table (for small k). The roll-your-own one I doubt is much
 * good, so you should probably use the MD5 hash. Notice
 * that machines send messages to each other that contain
 * the hash only, so all machines must use the same calculateKey
 * function.
 */
static long long unsigned
calculateKey(char *url)
{
  const static int useMD5 = 1;

  if(useMD5){
    return calculateMD5Key(url);
  }
  else{
    return calculateStupidKey(url);
  }

}


static long long unsigned
calculateMD5Key(char *url)
{
  MD5_CTX  tCrypt;
  static long long unsigned digestPart;

  assert(url);
  MD5Init(&tCrypt);
  MD5Update(&tCrypt, "this is my secret key, yeah right", strlen("this is my secret key, yeah right"));
  MD5Update(&tCrypt, url, strlen(url));
  MD5Final(&tCrypt);
  /*
   * The digest MUST be smaller than a long long (otherwise we
   * would be including undefined bytes past end of digest.
   */
  assert(MD5_DIGEST_SIZE * sizeof(unsigned char) 
	 >= sizeof(long long unsigned));
				   
  memcpy(&digestPart, &(tCrypt.digest[0]), sizeof(digestPart));
  return digestPart;
}

static long long unsigned
calculateStupidKey(char *url)
{
  static const long long unsigned BIG_PRIME_1 = 193844587;
  static const long long unsigned BIG_PRIME_2 = 455899;
  static const long long unsigned BIG_PRIME_3 = 901367;
  
  int len, ii;
  long long unsigned key, chunk;
  assert(url);
  len = strlen(url);
  key = 0;
  for(ii = 0; ii < len-8; ii += 8){
    /* Make copy to guarantee alignment. */
    memcpy(&chunk, &url[ii], 8);
    key = key + chunk * BIG_PRIME_1;
  }
  for( ; ii < len; ii++){
    key = key + (unsigned char)url[ii] * BIG_PRIME_2;
  }
  key = key + len * BIG_PRIME_3;
  return key;
}


void 
Net_URLKey_Init(Net_URLKey *mungedKey, URLKey *key)
{
  mungedKey->mungedKey = HCEndian_llMachineToNet(key->key);
}

void
Net_URLKey_LocalFormat(Net_URLKey *mungedKey, URLKey *key)
{
  key->key = HCEndian_llNetToMachine(mungedKey->mungedKey);
}

/*
 *------------------------------------------------------------------
 *
 * HCEntry --
 *
 *          A <URLKey, NodeId> pair. Generally points to
 *          the nearest copy of the specified url.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
void 
HCEntry_Init(HCEntry *entry, URLKey key, NodeId node)
{
  assert(entry);
  entry->key = key;
  entry->node = node;
}
void
HCEntry_Copy(HCEntry *dest, HCEntry *src)
{
  assert(dest);
  assert(src);
  dest->key = src->key;
  dest->node = src->node;
}
int
HCEntry_Compare(HCEntry *e1, HCEntry *e2)
{
  assert(e1);
  assert(e2);
  if(URLKey_Compare(&e1->key, &e2->key)){
    return 1;
  }
  if(NodeId_Compare(&e1->node, &e2->node)){
    return 1;
  }
  return 0;
}

void
Net_HCEntry_Init(Net_HCEntry *mungedEntry, HCEntry *entry)
{
  assert(mungedEntry);
  assert(entry);
  Net_URLKey_Init(&mungedEntry->mungedKey, &entry->key);
  Net_NodeId_Init(&mungedEntry->mungedNode, &entry->node);
}

void
Net_HCEntry_LocalFormat(Net_HCEntry *mungedEntry, HCEntry *entry)
{
  Net_URLKey_LocalFormat(&mungedEntry->mungedKey, &entry->key);
  Net_NodeId_LocalFormat(&mungedEntry->mungedNode, &entry->node);
}


/*
 *------------------------------------------------------------------
 *
 * HCUpdate --
 *
 *          An update is a HCEntry and an action.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
void
HCUpdate_Init(HCUpdate *u, enum HCUpdateEnum action, HCEntry *entry)
{
  assert(u);
  assert(entry);
  u->action = action;
  HCEntry_Copy(&u->entry, entry);
}
int
HCUpdate_Compare(HCUpdate *u1, HCUpdate *u2)
{
  assert(u1);
  assert(u2);
  if(u1->action != u2->action){
    return 1;
  }
  if(HCEntry_Compare(&u1->entry, &u2->entry)){
    return 1;
  }
  return 0;
}

void
HCUpdate_selfTest()
{
  assert(sizeof(enum HCUpdateEnum) == sizeof(long unsigned));
}

void
Net_HCUpdate_Init(Net_HCUpdate *mungedUpdate, HCUpdate *update)
{
  assert(mungedUpdate);
  assert(update);
  mungedUpdate->mungedAction = HCEndian_lMachineToNet(update->action);
  Net_HCEntry_Init(&mungedUpdate->mungedEntry, &update->entry);
}


void
Net_HCUpdate_LocalFormat(Net_HCUpdate *mungedUpdate, HCUpdate *update)
{
  assert(mungedUpdate);
  assert(update);
  update->action = HCEndian_lNetToMachine(mungedUpdate->mungedAction);
  Net_HCEntry_LocalFormat(&mungedUpdate->mungedEntry, &update->entry);
}

/*
 *------------------------------------------------------------------
 *
 * primitivesSelfTest --
 *
 *          description.
 *
 * Arguments:
 *      type1 arg1 -- description.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------
 */
void
primitivesSelfTest()
{

  NodeId nodeId, nodeId2;
  Net_NodeId netNodeId;
  URLKey urlKey, urlKey2;
  Net_URLKey netUrlKey;
  char url[128];
  HCEntry hcEntry, hcEntry2;
  Net_HCEntry netHCEntry;
  HCUpdate hcUpdate, hcUpdate2;
  Net_HCUpdate netHCUpdate;

  printf("primitivesSelfTest()..");

  NodeId_Init(&nodeId, 23423, 234);
  NodeId_selfTest(&nodeId);
  Net_NodeId_Init(&netNodeId, &nodeId);
  Net_NodeId_LocalFormat(&netNodeId, &nodeId2);
  assert(!NodeId_Compare(&nodeId, &nodeId2));

  sprintf(url, "http://www.cs.Utexas.edu");
  URLKey_Init(&urlKey, url);
  Net_URLKey_Init(&netUrlKey, &urlKey);
  Net_URLKey_LocalFormat(&netUrlKey, &urlKey2);
  assert(!URLKey_Compare(&urlKey, &urlKey2));

  HCEntry_Init(&hcEntry, urlKey, nodeId);
  Net_HCEntry_Init(&netHCEntry, &hcEntry);
  Net_HCEntry_LocalFormat(&netHCEntry, &hcEntry2);
  assert(!HCEntry_Compare(&hcEntry, &hcEntry2));

  HCUpdate_Init(&hcUpdate, HC_ChildInval, &hcEntry);
  HCUpdate_selfTest();
  Net_HCUpdate_Init(&netHCUpdate, &hcUpdate);
  Net_HCUpdate_LocalFormat(&netHCUpdate, &hcUpdate2);
  assert(!HCUpdate_Compare(&hcUpdate, &hcUpdate2));

  printf(".OK\n");
}
