/*
 * $Id: post.c,v 1.8 1998/05/11 15:38:07 dahlin Exp $
 *
 * DEBUG: section 52  POST transmission and reception
 * AUTHOR: Harvest Derived
 *
 * SQUID Internet Object Cache  http://squid.nlanr.net/Squid/
 * --------------------------------------------------------
 *
 *  Squid is the result of efforts by numerous individuals from the
 *  Internet community.  Development is led by Duane Wessels of the
 *  National Laboratory for Applied Network Research and funded by
 *  the National Science Foundation.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *  
 */

/*
 * Copyright (c) 1994, 1995.  All rights reserved.
 *  
 *   The Harvest software was developed by the Internet Research Task
 *   Force Research Group on Resource Discovery (IRTF-RD):
 *  
 *         Mic Bowman of Transarc Corporation.
 *         Peter Danzig of the University of Southern California.
 *         Darren R. Hardy of the University of Colorado at Boulder.
 *         Udi Manber of the University of Arizona.
 *         Michael F. Schwartz of the University of Colorado at Boulder.
 *         Duane Wessels of the University of Colorado at Boulder.
 *  
 *   This copyright notice applies to software in the Harvest
 *   ``src/'' directory only.  Users should consult the individual
 *   copyright notices in the ``components/'' subdirectories for
 *   copyright information about other software bundled with the
 *   Harvest source code distribution.
 *  
 * TERMS OF USE
 *   
 *   The Harvest software may be used and re-distributed without
 *   charge, provided that the software origin and research team are
 *   cited in any use of the system.  Most commonly this is
 *   accomplished by including a link to the Harvest Home Page
 *   (http://harvest.cs.colorado.edu/) from the query page of any
 *   Broker you deploy, as well as in the query result pages.  These
 *   links are generated automatically by the standard Broker
 *   software distribution.
 *   
 *   The Harvest software is provided ``as is'', without express or
 *   implied warranty, and with no support nor obligation to assist
 *   in its use, correction, modification or enhancement.  We assume
 *   no liability with respect to the infringement of copyrights,
 *   trade secrets, or any patents, and are not responsible for
 *   consequential damages.  Proper use of the Harvest software is
 *   entirely the responsibility of the user.
 *  
 * DERIVATIVE WORKS
 *  
 *   Users may make derivative works from the Harvest software, subject 
 *   to the following constraints:
 *  
 *     - You must include the above copyright notice and these 
 *       accompanying paragraphs in all forms of derivative works, 
 *       and any documentation and other materials related to such 
 *       distribution and use acknowledge that the software was 
 *       developed at the above institutions.
 *  
 *     - You must notify IRTF-RD regarding your distribution of 
 *       the derivative work.
 *  
 *     - You must clearly notify users that your are distributing 
 *       a modified version and not the original Harvest software.
 *  
 *     - Any derivative product is also subject to these copyright 
 *       and use restrictions.
 *  
 *   Note that the Harvest software is NOT in the public domain.  We
 *   retain copyright, as specified above.
 *  
 * HISTORY OF FREE SOFTWARE STATUS
 *  
 *   Originally we required sites to license the software in cases
 *   where they were going to build commercial products/services
 *   around Harvest.  In June 1995 we changed this policy.  We now
 *   allow people to use the core Harvest software (the code found in
 *   the Harvest ``src/'' directory) for free.  We made this change
 *   in the interest of encouraging the widest possible deployment of
 *   the technology.  The Harvest software is really a reference
 *   implementation of a set of protocols and formats, some of which
 *   we intend to standardize.  We encourage commercial
 *   re-implementations of code complying to this set of standards.  
 */

#include "squid.h"
#include <assert.h>

#define POST_PAGE_SIZE    (SM_PAGE_SIZE - sizeof(Poster))

#define POST_REPLY_HDR()     "HTTP/1.0 204 OK - no content\r\n"

#define POST_DELETE_GAP   (1<<18)

static void postProcessHeader _PARAMS((HttpStateData *, icpStateData *));
static void postHandleStore _PARAMS((int, StoreEntry *, Poster *));
static void postSendData _PARAMS((int, char *, int, int, Poster *));
static void postConnected _PARAMS((int fd, int commstat, Poster *));
static void postSendDone _PARAMS((int fd, Poster *p));
static peer *whichPostPeer _PARAMS((const struct sockaddr_in *from));


/* Authenticate by checking if other end is on local host or a parent */
int
postAuth(icpStateData *icpState)
{
    struct sockaddr_in peer;
    unsigned a;
    int plen;

    if (Config.Options.allow_posts == 0) {
	debug(52, 0, "warning: POST received but POSTs are disabled\n");
	return(0);
    }
    
    /* POSTS from neighbors are OK */
    plen = sizeof(peer);
    if (getpeername(icpState->fd, (struct sockaddr *) &peer, &plen) == -1) {
	debug(52, 0, "postAuth: getpeername on %d: error %d\n",
	      icpState->fd, errno);
    }
    else {
	/* POSTS from localhost are fine, too */
	a = ntohl(peer.sin_addr.s_addr);
	if (a == 0x7f000001)
	    /* From localhost, is ok */
	    return(1);
    }

    /* POSTS from peers are OK */
    return(whichPostPeer(&peer) != 0);
}

/*
 * Code for dealing with incoming POST requests
 */
void
postRecv(icpStateData *icpState)
{
    struct _http_reply *reply;
    HttpStateData *httpState;
    StoreEntry *entry;
    const char *pubkey;
    int off;
    
    /* Validate the request */
    if (!postAuth(icpState))
	return;

    pubkey = storeGeneratePublicKey(icpState->url, icpState->method);
    BIT_SET(icpState->request->flags, REQ_CACHABLE);
    icpState->log_type = LOG_POST;
    icpState->out_offset = 0;

    entry = storeGet(pubkey);
    
    if (entry)
	entry = storeReuse(entry);
    
    if (entry == NULL) {
	entry = storeCreateEntry(icpState->url, icpState->url,
				 icpState->request_hdr,
				 icpState->req_hdr_sz,
				 icpState->request->flags, icpState->method);
    }
    icpState->entry = entry;

    /* Need a request pointer in the mem_obj */
    requestLink(icpState->request);
    entry->mem_obj->request = icpState->request;
    
    debug(52, 4, "postRecv: %s for '%s'\n", log_tags[icpState->log_type],
	  icpState->url);

    /* Create an HTTP context from ICP info */
    /* Should factor out common stuff between this and proxyHttpStart */
    httpState = httpFromIcp(icpState);
    httpState->entry = entry;
    comm_add_close_handler(icpState->fd, httpStateFree, httpState);

    /* Prepare to reply if POST succeeds. */
    comm_add_close_handler(icpState->fd, (PF) postReply, entry);
    
    /* Parse the headers */
    postProcessHeader(httpState, icpState);
    reply = entry->mem_obj->reply;

    /* Send ROUTEs to hint cache */
    if (entry->mem_obj->request->protocol == PROTO_ROUTE) {
	HCRecv(icpState);
    }
    else {
        /* Update the timestamp */
        if (entry->mem_obj->reply->last_modified)
	    entry->lastmod = entry->mem_obj->reply->last_modified;
	else
	    /* Make it clear the version has changed even if no lastmod */
	    ++entry->lastmod;
	
	/* Distribute this update */
	distEntryUpdate(entry, icpState);
    }

    /* httpReadReply reads msgs with bodies */
    off = storeGetLowestReaderOffset(entry);
    if (off < reply->content_length + reply->hdr_sz){
      httpReadReply(icpState->fd, httpState);
    }
    else{
      postReplyAndClose(icpState->fd, entry);
    }
}

static peer *
whichPostPeer(const struct sockaddr_in *from)
{
    int j;
    u_short port = ntohs(from->sin_port);
    struct in_addr ip = from->sin_addr;
    peer *e = NULL;
    debug(15, 3, "whichPostPeer: from %s port %d\n", inet_ntoa(ip), port);
    for (e = getFirstPeer(); e; e = getNextPeer(e)) {
	for (j = 0; j < e->n_addresses; j++) {
	  /*
	   * XXX Should really be something like: 
	   * if (ip.s_addr == e->addresses[j].s_addr) && port == e->icp_port) {
	   * but the remote port number is essentially random (since
	   * we're the server for this request), and it doesn't 
	   * provide any authentication information. The real fix is
	   * to put in the real encryption-based authentication.
	   */
	    if (ip.s_addr == e->addresses[j].s_addr){ 
		return e;
	    }
	}
    }
    return NULL;
}

/* Process POST header. */
static void
postProcessHeader(HttpStateData *httpState, icpStateData *icpState)
{
    static char replline[] = "HTTP/1.0 200 OK";
    char *hdr = icpState->inbuf;
    char *eol;

    /* First change the POST first line to a reply first line */
    eol = strchr(hdr, '\r');
    eol[0] = '\0';
    strcpy(hdr, replline);
    memset(hdr + sizeof(replline) - 1, ' ', eol - hdr - sizeof(replline) + 1);
    eol[0] = '\r';

    /* Make request look like a reply */
    /* XXX - this probably screws up logging */
    httpState->request->method = METHOD_GET;
    
    httpProcessReplyHeader(httpState, hdr, icpState->in_offset);
    storeAppend(icpState->entry, hdr, icpState->in_offset);
}

#if 0
static void
postReplyDone(int fd, char *buf, int off, int status, void *data)
{
    comm_close(fd);
}
#endif

void
postReplyAndClose(int fd, StoreEntry *entry)
{
  postReply(fd, entry);
  comm_close(fd);
}

void
postReply(int fd, StoreEntry *entry)
{
  int sent, tot = 0;

    if (entry->mem_obj->reply->code != 200)
	/* Already aborted, hopefully with error msg */
	return;

    /*
     * POST succeeded, acknowledge this.
     * Note: comm_write is inappropriate here because this
     * is a close handler.  Using bare write() is a bit of a hack;
     * it is 99.9% likely to work because this is the only thing
     * we're writing back and it is of strictly limited length.
     * In the remaining case, the "failure" is that we block, which
     * will hurt performance, but still operate correctly.
     */
#if 0
    comm_write(fd, POST_REPLY_HDR(), sizeof(POST_REPLY_HDR()),
	       0, (rw_complete_handler *) postReplyDone, 0, 0);
    return;
#else
    while(tot < sizeof(POST_REPLY_HDR())){
      sent = write(fd, POST_REPLY_HDR(), sizeof(POST_REPLY_HDR()));
      if(sent < 0){
	/* ERROR -- not much we can do. We're closing this fd anyhow */
	return;
      }
      tot += sent;
    }
    return;
#endif
}


/* Output POSTs */

/* Make up and send a POST to a given location */
void
postSend(StoreEntry *entry, struct sockaddr_in *saddr, PF handler, void *arg)
{
    struct in_addr nobody;
    Poster *p;
    char *buf;
    int fd;

    /* Make sure entry stays inmem until operation complete */
    if (storeLockObject(entry, NULL, NULL) < 0) {
	debug(52, 1, "postSend: cannot lock object.\n");
	assert(0);  /* XXX! */
	return;
    }
	    
    /* Need to create and connect a POST socket */
    nobody.s_addr = INADDR_ANY;
    fd = comm_open(SOCK_STREAM, 0, nobody, 0, COMM_NONBLOCKING, "tcppost");
    if (fd == COMM_ERROR) {
	debug(11, 1, "postSend: Could not create new socket.\n");
	storeUnlockObject(entry);
	return;
    }

    /* Allocate memory for buffer and state tracking */
    p = (Poster *) get_free_4k_page();
    buf = (char *) &p[1];
    p->flags = P_CONNECTING;
    p->entry = entry;
    p->handler = handler;
    p->arg = arg;

    /* Deal with socket closing */
    comm_add_close_handler(fd, (PF) postSendDone, entry);
    
    /* Connect to this updatee */
    strcpy(p->hname, inet_ntoa(saddr->sin_addr));
    commConnectStart(fd, p->hname, saddr->sin_port, (CCH) postConnected, p);
}

/* Send the object's data */
static void
postSendDone(int fd, Poster *p)
{
    StoreEntry *entry = p->entry;

    if (!(p->flags & P_ACTIVE))
	return;
    p->flags &= ~P_ACTIVE;

    if (p->handler)
	(*p->handler)(fd, p->arg);	
    
    storeUnregister(entry, fd);
    storeUnlockObject(entry);
    
    /* Free 4k buffer and poster state along with it */
    xfree(p);
    
    /* For now, close when have nothing to send - get smarter later */
    comm_close(fd);
}

/* xlate storeRegister calling conventions to comm_write ones */
static void
postHandleStore(int fd, StoreEntry *entry, Poster *p)
{
    /* Reregister immediately so store has client ds for fd */
    storeRegister(entry, fd, (PIF) postHandleStore, p);
    
    /* Do the work. */
    postSendData(fd, get_free_4k_page(), 0, 0, p);
}

/* Xfer data from mem_obj to updatee */
/* XXX - duplication with icpSendMoreData, etc. */
static void
postSendData(int fd, char *buf, int size, int errflag, Poster *p)
{
    StoreEntry *entry = p->entry;
    int len;
    
    debug(53, 5, "postSendData: FD %d '%s'\n", fd, entry->url);
    
    if (errflag) {
	debug(53, 3, "postSendData: write failed, errno %d\n", errno);
	postSendDone(fd, p);
	return;
    }
    
    if (size > 0)
	/* We must have been called by successful comm_write */
	p->flags &= ~P_WRITING;

    if (!(p->flags & P_CONNECTED) || (p->flags & P_WRITING))
	return;
    
    p->off += size;
    if (p->off >= entry->mem_obj->e_current_len) {
	/* We've sent everything we've seen so far */
	put_free_4k_page(buf);
	if (entry->object_len == entry->mem_obj->e_current_len)
	    /* We've seen all we're gonna see of this version */
	    postSendDone(fd, p);
	return;
    }
    
    storeClientCopy(entry, p->off, POST_PAGE_SIZE, buf, &len, fd);
    p->flags |= P_WRITING;
    comm_write(fd, buf,	len, 30, (rw_complete_handler *) postSendData, p, 0);
}

/* Update given updatee on contents of entry */
static void
postConnected(int fd, int commstat, Poster *p)
{
    StoreEntry *entry = p->entry;
    char *buf;
    
    if (commstat == COMM_ERROR) {
	/* XXX - this should be more sophisticated */
	debug(11, 4, "postConnected: Could not connect to %s\n", p->hname);
	comm_close(fd);
	return;
    }
    
    p->flags &= ~P_CONNECTING;
    p->flags |= P_CONNECTED;
    
    /* Compose header */
    buf = get_free_4k_page();
    sprintf(buf, "POST %s HTTP/1.0\r\n", entry->url);
    p->off = -strlen(buf);
    
    /* Write object asynchronously, starting with header */
    storeRegister(entry, fd, (PIF) postHandleStore, p);
    comm_write(fd, buf, strlen(buf), 30,
	       (rw_complete_handler *) postSendData, p, 0);
}

